import numpy as np,sys,os
from scipy.interpolate import interp1d, RectBivariateSpline
from . import grids_utils

def get_rhop_RZ(R,Z, geqdsk):
    '''Find rhop at every R,Z [m] based on the equilibrium in the geqdsk dictionary.
    '''
    return RectBivariateSpline(geqdsk['AuxQuantities']['Z'],
                               geqdsk['AuxQuantities']['R'],
                               geqdsk['AuxQuantities']['RHOpRZ']).ev(Z,R)



def vol_average(quant, rhop, method='omfit', geqdsk=None, device=None, shot=None, time=None, return_geqdsk=False):
    '''Calculate the volume average of the given radially-dependent quantity on a rhop grid. 

    Parameters
    ----------
    quant : array, (space, ...)
        quantity that one wishes to volume-average. The first dimension must correspond to space,
        but other dimensions may be exist afterwards.
    rhop : array, (space,)
        Radial rhop coordinate in cm units. 
    method : {'omfit','fs'}
        Method to evaluate the volume average. The two options correspond to the way to compute
        volume averages via the OMFIT fluxSurfaces classes and via a simpler cumulative sum in r_V 
        coordinates. The methods only slightly differ in their results. Note that 'omfit' will fail if 
        rhop extends beyond the LCFS, while method 'fs' can estimate volume averages also into the SOL.
        Default is method='omfit'. 
    geqdsk : output of the :py:class:`omfit_classes.omfit_eqdsk.OMFITgeqdsk` class, postprocessing the EFIT geqdsk file
        containing the magnetic geometry. If this is left to None, the function internally tries to fetch
        it using MDS+ and `omfit_classes.omfit_eqdsk`. In this case, device, shot and time to fetch the equilibrium 
        are required. 
    device : str
        Device name. Note that routines for this device must be implemented in `omfit_classes.omfit_eqdsk` for this to work. 
    shot : int
        Shot number of the above device, e.g. 1101014019 for C-Mod.
    time : float
        Time at which equilibrium should be fetched in units of ms. 
    return_geqdsk : bool
        If True, `omfit_classes.omfit_eqdsk` dictionary is also returned

    Returns
    -------
    quant_vol_avg : array, (space, ...)
        Volume average of the quantity given as an input, in the same units as in the input.
        If extrapolation beyond the range available from EFIT volume averages over a shorter section
        of the radial grid will be attempted. This does not affect volume averages within the LCFS. 
    geqdsk : dict
        Only returned if return_geqdsk=True. 
    '''
    if time is not None and np.mean(time)<1e2:
        print('Input time was likely in the wrong units! It should be in [ms]!')
    if np.max(rhop)>1.0:
        print("Input rhop goes beyond the LCFS! Results may not be meaningful (and can only be obtained via method=='fs').")
        
    if geqdsk is None:
        # Fetch device geqdsk from MDS+ and post-process it using the OMFIT geqdsk format.
        try:
            from omfit_classes import omfit_eqdsk
        except:
            raise ValueError('Could not import omfit_classes.omfit_eqdsk! Install with pip install omfit_classes')
        geqdsk = omfit_eqdsk.OMFITgeqdsk('').from_mdsplus(
            device=device, shot=shot, time=time, SNAPfile='EFIT01',
            fail_if_out_of_range=False,time_diff_warning_threshold=20)

    if method=='fs':
        # obtain mapping between rhop and r_V coordinates
        rho_pol, r_V_ = grids_utils.get_rhopol_rvol_mapping(geqdsk)
        
        # find r_V corresponding to input rhop (NB: extrapolation in the far SOL should be used carefully)
        r_V = interp1d(rho_pol, r_V_, bounds_error=False)(rhop)
        
        # use convenient volume averaging in r_V coordinates
        if np.any(np.isnan(r_V)):
            print('Ignoring all nan points! These may derive from an attempted extrapolation or from nan inputs')
            
        vol_avg = rV_vol_average(quant[~np.isnan(r_V)], r_V[~np.isnan(r_V)])
        
    elif method=='omfit':
        # use fluxSurfaces classes from OMFIT
        rhopp = np.sqrt(geqdsk['fluxSurfaces']['geo']['psin'])
        quantp = interp1d(rhop, quant, bounds_error=False,fill_value='extrapolate')(rhopp)
        vol_avg = geqdsk['fluxSurfaces'].volume_integral(quantp)

    else:
        raise ValueError('Input method for volume average could not be recognized')

    if return_geqdsk:
        return vol_avg, geqdsk
    else:
        return vol_avg


def rV_vol_average(quant,r_V):
    '''Calculate a volume average of the given radially-dependent quantity on a r_V grid.
    This function makes useof the fact that the r_V radial coordinate, defined as 
    :math:`r_V = \sqrt{ V / (2 \pi^2 R_{axis} }`,
    maps shaped volumes onto a circular geometry, making volume averaging a trivial 
    operation via
    :math:`\langle Q \rangle = \Sigma_i Q(r_i) 2 \pi \ \Delta r_V`
    where :math:`\Delta r_V` is the spacing between radial points in r_V.
    
    Note that if the input r_V coordinate is extended outside the LCFS,
    this function will return the effective volume average also in the SOL, since it is 
    agnostic to the presence of the LCFS. 

    Parameters
    ----------
    quant : array, (space, ...)
        quantity that one wishes to volume-average. The first dimension must correspond to r_V,
        but other dimensions may be exist afterwards. 
    r_V : array, (space,)
        Radial r_V coordinate in cm units. 

    Returns
    -------
    quant_vol_avg : array, (space, ...)
        Volume average of the quantity given as an input, in the same units as in the input
    '''
    quant_vol_avg = 2.*np.cumsum(quant*r_V*np.diff(r_V,prepend=0.0)) / (r_V[-1]**2 )

    return quant_vol_avg


    
def rad_coord_transform(x,name_in,name_out, geqdsk):
    """Transform from one radial coordinate to another. Note that this coordinate conversion is only
    strictly valid inside of the LCFS.

    Parameters
    ----------
    x: array or float
        input x coordinate
    name_in: str
        input x coordinate name ('rhon','r_V','rhop','rhov','Rmid','rmid','r/a')
    name_out: str
        input x coordinate ('rhon','psin','r_V', 'rhop','rhov','Rmid','rmid','r/a')
    geqdsk: dict
        gEQDSK dictionary, as obtained from the omfit-eqdsk package. 
    
    Returns
    -------
    array or float
        Conversion of `x` input for the requested radial grid coordinate.
    """
    if name_in == name_out:
        return x

    if 'r_V' not in geqdsk['fluxSurfaces']['geo']:
        R0 = geqdsk['RMAXIS']
        eq_vol = geqdsk['fluxSurfaces']['geo']['vol']
        r_V = np.sqrt(eq_vol/(2*np.pi**2*R0))
        geqdsk['fluxSurfaces']['geo']['r_V'] = r_V

    # sqrt(norm. tor. flux)
    rhon_ref = geqdsk['fluxSurfaces']['geo']['rhon']
    # norm. pol. flux
    psin_ref = geqdsk['fluxSurfaces']['geo']['psin']
    # sqrt(norm. pol. flux)
    rhop_ref = np.sqrt(psin_ref)
    #volume radius
    r_V = geqdsk['fluxSurfaces']['geo']['r_V']
    # R at midplane
    Rmid = geqdsk['fluxSurfaces']['midplane']['R']
    # r at midplane
    R0 = geqdsk['fluxSurfaces']['R0']
    rmid = Rmid - R0

    # Interpolate to transform coordiantes
    if name_in == 'rhon':
        coord_in = rhon_ref
    elif name_in == 'rhop':
        coord_in = rhop_ref
    elif name_in == 'r_V':
        coord_in = r_V
    elif name_in == 'rhov':
        r_V_lcfs = np.interp(1, rhon_ref, r_V)
        coord_in = r_V/r_V_lcfs
    elif name_in == 'Rmid':
        coord_in = Rmid
    elif name_in == 'rmid':
        coord_in = rmid
    elif name_in == 'r/a':
        rmid_lcfs = np.interp(1, rhon_ref, rmid)
        coord_in = rmid/rmid_lcfs
    else:
        raise ValueError('Input coordinate was not recognized!')

    if name_out == 'rhon':
        coord_out = rhon_ref
    elif name_out == 'psin':
        coord_out = psin_ref
    elif name_out == 'rhop':
        coord_out = rhop_ref
    elif name_out == 'r_V':
        coord_out = r_V
    elif name_out == 'rhov':
        r_V_lcfs = np.interp(1, rhon_ref, r_V)
        coord_out = r_V/r_V_lcfs
    elif name_out == 'Rmid':
        coord_out = Rmid
    elif name_out == 'rmid':
        coord_out = rmid
    elif name_out == 'r/a':
        rmid_lcfs = np.interp(1, rhon_ref, rmid)
        coord_out = rmid/rmid_lcfs
    else:
        raise ValueError('Output coordinate was not recognized!')

    ind = coord_in != 0

    if isinstance(x, (float,int)) and float(x)==0.:
        # interpolation trick below doesn't play nicely with magnetic axis
        return np.interp(x,coord_in[ind],coord_out[ind])
    else:
        #trick for better extrapolation
        return np.interp(x,coord_in[ind],coord_out[ind]/coord_in[ind])*x

        




import matplotlib.pyplot as plt
import numpy as np, copy
import matplotlib.gridspec as mplgs
import matplotlib.widgets as mplw
from matplotlib.cm import ScalarMappable
import itertools
plt.ion()


def slider_plot(x, y, z, xlabel='', ylabel='', zlabel='', labels=None, plot_sum=False,
                x_line=None, y_line=None, **kwargs):
    """Make a plot to explore multidimensional data.

    Parameters
    ----------
    x : array of float, (`M`,)
        The abscissa. (in aurora, often this may be rhop)
    y : array of float, (`N`,)
        The variable to slide over. (in aurora, often this may be time)
    z : array of float, (`P`, `M`, `N`)
        The variables to plot.
    xlabel : str, optional
        The label for the abscissa.
    ylabel : str, optional
        The label for the slider.
    zlabel : str, optional
        The label for the ordinate.
    labels : list of str with length `P`
        The labels for each curve in `z`.
    plot_sum : bool, optional
        If True, will also plot the sum over all `P` cases. Default is False.
    x_line : float, optional
        x coordinate at which a vertical line will be drawn. 
    y_line : float, optional
        y coordinate at which a horizontal line will be drawn.
    """
    if labels is None:
        labels = ['' for v in z]

    # make sure not to modify the z array in place
    zz = copy.deepcopy(z)

    fig = plt.figure()
    fig.set_size_inches(10,7, forward=True)
    
    # separate plot into 3 subgrids
    a_plot = plt.subplot2grid((10,10),(0,0),rowspan = 8, colspan = 8, fig=fig) 
    a_legend = plt.subplot2grid((10,10),(0,8),rowspan = 8, colspan = 2, fig=fig) 
    a_slider = plt.subplot2grid((10,10),(9,0),rowspan = 1, colspan = 8, fig=fig) 
    
    a_plot.set_xlabel(xlabel)
    a_plot.set_ylabel(zlabel)
    if x_line is not None: a_plot.axvline(x_line, c='r',ls=':',lw=0.5)
    if y_line is not None: a_plot.axhline(y_line, c='r',ls=':',lw=0.5)

    ls_cycle = get_ls_cycle()

    l = []

    # plot all lines
    for v, l_ in zip(zz, labels):
        ls = next(ls_cycle)
        tmp, = a_plot.plot(x, v[:, 0], ls, **kwargs)
        _ = a_legend.plot([], [], ls, label=l_, **kwargs)
        l.append(tmp)

    if plot_sum:
        # add sum of the first axis to the plot (and legend)
        ls = next(ls_cycle)
        l_sum, = a_plot.plot(x, zz[:, :, 0].sum(axis=0), ls, **kwargs)
        _ = a_legend.plot([],[], ls, label='total', **kwargs)

    leg=a_legend.legend(loc='best', fontsize=12).set_draggable(True)
    title = fig.suptitle('')
    a_legend.axis('off')
    a_slider.axis('off')


    def update(dum):

        i = int(slider.val)

        for v, l_ in zip(zz, l):
            l_.set_ydata(v[:, i])

        if plot_sum:
            l_sum.set_ydata(zz[:, :, i].sum(axis=0))

        a_plot.relim()
        a_plot.autoscale()

        title.set_text('%s = %.5f' % (ylabel, y[i]) if ylabel else '%.5f' % (y[i],))

        fig.canvas.draw()

    def arrow_respond(slider, event):
        if event.key == 'right':
            slider.set_val(min(slider.val + 1, slider.valmax))
        elif event.key == 'left':
            slider.set_val(max(slider.val - 1, slider.valmin))

    slider = mplw.Slider(
        a_slider,
        ylabel,
        0,
        len(y) - 1,
        valinit=0,
        valfmt='%d'
    )
    slider.on_changed(update)
    update(0)
    fig.canvas.mpl_connect(
        'key_press_event',
        lambda evt: arrow_respond(slider, evt)
    )




def get_ls_cycle():
    color_vals = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
    style_vals = ['-', '--', '-.', ':']
    ls_vals = []
    for s in style_vals:
        for c in color_vals:
            ls_vals.append(c + s)
    return itertools.cycle(ls_vals)


def get_color_cycle():
    color_vals = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
    return itertools.cycle(color_vals)


def get_line_cycle():
    style_vals = ['-', '--', '-.', ':']
    return itertools.cycle(style_vals)


class DraggableColorbar:
    '''Create a draggable colorbar for matplotlib plots to enable quick changes in color scale. 

    Example:::

        fig,ax = plt.subplots()
        cntr = ax.contourf(R, Z, vals)
        cbar = plt.colorbar(cntr, format='%.3g', ax=ax)
        cbar = DraggableColorbar(cbar,cntr)
        cbar.connect()
    '''
    def __init__(self, cbar, mapimage):
        self.cbar = cbar
        self.mapimage = mapimage
        self.press = None
        self.cycle = sorted([i for i in dir(plt.cm) if hasattr(getattr(plt.cm,i),'N')])

        self.index = self.cycle.index(ScalarMappable.get_cmap(cbar).name)

    def connect(self):
        '''Matplotlib connection for button and key pressing, release, and motion.
        '''
        self.cidpress = self.cbar.patch.figure.canvas.mpl_connect('button_press_event', self.on_press)
        self.cidrelease = self.cbar.patch.figure.canvas.mpl_connect('button_release_event', self.on_release)
        self.cidmotion = self.cbar.patch.figure.canvas.mpl_connect('motion_notify_event', self.on_motion)
        self.keypress = self.cbar.patch.figure.canvas.mpl_connect('key_press_event', self.key_press)

    def on_press(self, event):
        '''Button pressing; check if mouse is over colorbar.
        '''
        if event.inaxes != self.cbar.ax: return
        self.press = event.x, event.y

    def key_press(self, event):
        '''Key pressing event
        '''
        if event.key=='down':
            self.index += 1
        elif event.key=='up':
            self.index -= 1

        if self.index<0:
            self.index = len(self.cycle)
        elif self.index>=len(self.cycle):
            self.index = 0

        cmap = self.cycle[self.index]

        self.cbar.set_cmap(cmap)
        self.cbar.draw_all()
        self.mapimage.set_cmap(cmap)
        self.mapimage.get_axes().set_title(cmap)
        self.cbar.patch.figure.canvas.draw()

    def on_motion(self, event):
        '''Move if the mouse is over the colorbar.
        '''
        if self.press is None: return
        if event.inaxes != self.cbar.ax: return
        xprev, yprev = self.press
        dx = event.x - xprev
        dy = event.y - yprev
        self.press = event.x,event.y

        scale = self.cbar.norm.vmax - self.cbar.norm.vmin
        perc = 0.03
        if event.button==1:
            self.cbar.norm.vmin -= (perc*scale)*np.sign(dy)
            self.cbar.norm.vmax -= (perc*scale)*np.sign(dy)
        elif event.button==3:
            self.cbar.norm.vmin -= (perc*scale)*np.sign(dy)
            self.cbar.norm.vmax += (perc*scale)*np.sign(dy)
        self.cbar.draw_all()
        self.mapimage.set_norm(self.cbar.norm)
        self.cbar.patch.figure.canvas.draw()


    def on_release(self, event):
        '''Upon release, reset press data
        '''
        self.press = None
        self.mapimage.set_norm(self.cbar.norm)
        self.cbar.patch.figure.canvas.draw()

    def disconnect(self):
        self.cbar.patch.figure.canvas.mpl_disconnect(self.cidpress)
        self.cbar.patch.figure.canvas.mpl_disconnect(self.cidrelease)
        self.cbar.patch.figure.canvas.mpl_disconnect(self.cidmotion)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os, sys
sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('../aurora'))
#autodoc_mock_imports = ["aurora"]


# -- Project information -----------------------------------------------------

project = 'Aurora'
copyright = '2020, F.Sciortino'
author = 'F.Sciortino'

# The full version, including alpha/beta/rc tags
release = '1.0.0'

# FS: this seems to be useful for local build
master_doc = 'index'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.autosectionlabel',
              'sphinx.ext.todo',
              'sphinx.ext.mathjax',  #
              #'sphinx.ext.imgmath',
              'sphinx.ext.viewcode',
              'sphinx.ext.napoleon',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
on_rtd = os.environ.get('READTHEDOCS',None) == 'True'

if not on_rtd: # only import and set the theme if we're buiding docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

#html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = [] #['_static']




# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# avoid empty pages:
'extraclassoptions': 'openany,oneside',

# The font size ('10pt', '11pt' or '12pt').
'pointsize': '11pt',

# Additional stuff for the LaTeX preamble.
'preamble': r'\hypersetup{bookmarksdepth=3}',
}


# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'aurora.tex', u'Aurora Documentation',
   u'Francesco Sciortino', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'Aurora', u'Aurora Documentation',
     [u'Francesco Sciortino'], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'Aurora', u'Aurora Documentation',
   u'Francesco Sciortino', 'Aurora', 'Modern 1.5D impurity transport and radiation modeling for magnetically-confined plasmas.', 'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'



''' This script contains a number of functions used for interpolation of kinetic profiles and D,V profiles in STRAHL.
Refer to the STRAHL manual for details.
'''
from scipy.interpolate import interp1d
import numpy as np


def funct(params,rLCFS,r):
    ''' Function 'funct' in STRAHL manual

    The "params" input is broken down into 6 arguments:
        y0 is core offset
        y1 is edge offset
        y2 (>y0, >y1) sets the gaussian amplification
        p0 sets the width of the inner gaussian
        P1 sets the width of the outer gaussian
        p2 sets the location of the inner and outer peaks
    '''
    params = np.reshape(params,(-1,6))
    out = []
    for param in params:
        y0,y1,y2,p0,p1,p2 = param
        r1 = p2*rLCFS

        rin = r[r<=r1]
        rout = r[r>r1]

        yin  = y0 + (y2-y0) * np.exp(np.maximum(-(rin -r1)**2/p0**2,-50))
        yout = y1 + (y2-y1) * np.exp(np.maximum(-(rout-r1)**2/p1**2,-50))
        out.append(np.concatenate((yin,yout)))

    return np.array(out)


def funct2(params,rLCFS,r):
    '''Function 'funct2' in STRAHL manual.   
    '''
    params_1,params_2 = np.swapaxes(np.reshape(params,(-1,2,6)),0,1)

    funct_1 = funct(params_1,rLCFS,r)
    funct_2 = funct(params_2,rLCFS,r)
    return funct_1+funct_2


def exppol0(params,d,rLCFS,r):
    rho = r[:,None]/rLCFS
    d = np.array(d)/rLCFS
    params = np.array(params).T
    idx = np.searchsorted(r,rLCFS)
    core = params[0]*np.exp(params[1]*rho[:idx]**2 +
                            params[2]*rho[:idx]**4 +
                            params[3]*rho[:idx]**6 +
                            params[4]*rho[:idx]**8)
    edge = core[-1]*np.exp(-(rho[idx:]-rho[idx-1])/d)
    return np.concatenate([core,edge]).T

def exppol1(params,d,rLCFS,r):
    rho = r[:,None]/rLCFS
    d = np.array(d)/rLCFS
    params = np.array(params).T
    idx = np.searchsorted(r,rLCFS)
    core = params[0]*np.exp(params[1]*rho[:idx]**2 +
                            params[2]*rho[:idx]**3 +
                            params[3]*rho[:idx]**4 +
                            params[4]*rho[:idx]**5)
    edge = core[-1]*np.exp(-(rho[idx:]-rho[idx-1])/d)
    return np.concatenate([core,edge]).T


def ratfun(params,d,rLCFS,r):
    rho = r[:,None]/rLCFS
    d = np.array(d)/rLCFS
    params = np.array(params).T
    idx = np.searchsorted(r,rLCFS)
    core = params[0] * ( (1.0 - params[1])*(1.0-rho[:idx]**params[2])**params[3] + params[1])
    edge = core[-1]*np.exp(-(rho[idx:]-rho[idx-1])/d)
    return np.concatenate([core,edge]).T

def interp_quad(x,y,d,rLCFS,r):
    '''Function 'interp' used for kinetic profiles.
    '''
    f = interp1d(x,np.log(y), kind='quadratic', assume_sorted=True, copy=False)
    idx = np.searchsorted(r,rLCFS)
    core = np.exp(f(np.clip(r[:idx]/rLCFS,0, x[-1])))
    edge =  core[...,[idx-1]]*np.exp(-np.outer(1./np.asarray(d),r[idx:]-r[idx-1]))

    return np.concatenate([core,edge],axis=-1)

def interpa_quad(x,y,rLCFS,r):
    '''Function 'interpa' used for kinetic profiles
    '''
    f = interp1d(x,np.log(y),bounds_error=False, kind='quadratic',
                             assume_sorted=True, copy=False)
    return np.exp(f(np.minimum(r/rLCFS, x[-1])))


def interp(x,y,rLCFS,r):
    '''Function 'interp' used in STRAHL for D and V profiles.
    '''
    f = interp1d(x,y,fill_value='extrapolate', assume_sorted=True, copy=False)
    return f(r/rLCFS)



'''Functions to provide default ADAS files for Aurora modelling, including capabilities to fetch
these files remotely from the OPEN-ADAS website.
'''

import urllib
import shutil,os

# location of the "adas_data" directory relative to this script:
adas_data_dir = os.path.dirname(os.path.realpath(__file__))+os.sep+'adas_data'+os.sep

def get_adas_file_loc(filename, filetype='adf11'):
    '''Find location of requested atomic data file for the indicated ion. 
    The search proceeds with the following attempts, in this order:

    #. If the file is available in Aurora/adas_data/*filetype*, with filetype given by the user, 
       always use this data.

    #. If the input filename is actually a full path to the file on the local system, use this file 
       and copy it to Aurora/aurora/adas_data/actual_filename, where actual_filename is the file
       name rather than the full path. 

    #. If the environmental variable "AURORA_ADAS_DIR" is defined, attempt to find the file there, 
       under adf11/ and adf15/ directories. AURORA_ADAS_DIR may for example be the path to a central
       repository of ADAS files for Aurora on a cluster where not everyone may have write-permissions.
       For this option, files are not copied at all.

    #. Attempt to fetch the file remotely via open.adas.ac.uk and save it in 
       Aurora/aurora/adas_data/*filetype*/. 

    Parameters
    ----------
    filename : str
        Name of the ADAS file of interest, e.g. 'plt89_ar.dat'.
    filetype : str
        ADAS file type. Currently allowed: 'adf11' or 'adf15'
    
    Returns
    -------
    file_loc : str
        Full path to the requested file. 
    '''
    def fetch_file(filename,filetype, loc):
        if not os.path.isdir(os.path.dirname(os.path.dirname(loc))):
            # make sure that aurora/adas_data/adf** directory exists, sp that we can store data files in it
            os.makedirs(os.path.dirname(os.path.dirname(loc)))
        if not os.path.isdir(os.path.dirname(loc)):
            # make sure that aurora/adas_data/adf** directory exists, sp that we can store data files in it
            os.makedirs(os.path.dirname(loc))

        if filetype=='adf11':
            fetch_adf11_file(filename, loc)
        elif filetype=='adf15':
            fetch_adf15_file(filename, loc)
        else:
            raise ValueError('ADAS file type/format not recognized. Could not find it or download it automatically!')

    if os.path.exists(adas_data_dir+filetype+os.sep+filename):
        # file is available in adas_data:
        return adas_data_dir+filetype+os.sep+filename

    elif os.path.exists(filename):
        # check if user actually gave a complete filepath. Don't copy the file from the original location
        filepath = filename
        filename = filepath.split('/')[-1]
        #shutil.copyfile(filepath, adas_data_dir+filetype+os.sep+filename)
        return adas_data_dir+filetype+os.sep+filename
    
    elif 'AURORA_ADAS_DIR' in os.environ:
        loc = os.environ['AURORA_ADAS_DIR']+os.sep+filetype+os.sep+filename
        if os.path.exists(loc):
            pass
        else:
            # File could not be found. Download it and save it in AURORA_ADAS_DIR:
            fetch_file(filename, filetype, loc=loc)
        return loc

    else:
        # File could not be found. Download it and save it in adas_data:
        if not os.path.isdir(adas_data_dir):
            # make sure that aurora/adas_data exists
            os.makedirs(adas_data_dir)

        loc = adas_data_dir+filetype+os.sep+filename
        fetch_file(filename,filetype, loc=loc)
        return loc

    


def fetch_adf11_file(filename, loc):
    '''Download ADF11 file from the OPEN-ADAS website and store it in the 'adas_data/adf11'
    directory. 

    Parameters
    ----------
    filename : str
        Name of ADF11 file to be downloaded, e.g. 'plt89_ar.dat'.
    loc : str
        Location to save fetched ADF11 in.
    '''
    url = 'https://open.adas.ac.uk/download/adf11/'
    str1 =  filename.split('_')[0]
    
    local_filename,headers = urllib.request.urlretrieve(
        url+str1+'/'+filename,
        filename = loc
    )



def fetch_adf15_file(filename, loc):
    '''Download ADF15 file from the OPEN-ADAS website and store it in the 'adas_data/adf15'
    directory. 

    Parameters
    ----------
    filename : str
        Name of ADF15 file to be downloaded, e.g. 'pec96#c_pju#c2.dat'.
    loc : str
        Location to save fetched ADF15 file in.
    '''
    url = 'https://open.adas.ac.uk/download/adf15/'

    if filename.startswith('pec'):
        # more standard format, the following patterns should work fine:
        num = filename[3:5]
        spec = filename.split('#')[1].split('_')[0]
        filename_mod = filename.replace('#','][').replace('_',f'/pec{num}][{spec}_')
    elif filename.startswith('transport'):
        # different link format for files with "transport" name:
        filename_mod = 'transport/'+filename.replace('#','][')
    else:
        # patterns may be different, attempt simple guess:
        filename_mod = filename.split('_')[0]+'/'+filename.replace('#','][')

    local_filename,headers = urllib.request.urlretrieve(
        url+'/'+filename_mod,
        filename=loc
    )
    

def adas_files_dict():
    '''Selections for ADAS files for Aurora runs and radiation calculations.
    This function can be called to fetch a set of default files, which can then be modified (e.g. to 
    use a new file for a specific SXR filter) before running a calculation. 

    Returns
    -------
    files : dict
        Dictionary with keys equal to the atomic symbols of many of the most common ions of
        interest in fusion research. For each ion, a sub-dictionary contains recommended file 
        names for the relevant ADAS data types. Not all files types are available for all ions. 
        Files types are usually a subset of 
        'acd','scd','prb','plt','ccd','prc','pls','prs','fis','brs','pbs',prc'
        Refer to :py:func:`~aurora.atomic.get_adas_file_types` for a description of the meaning of 
        each file.
    '''
            
    files={}
    files["H"] = {}   #1
    files["H"]['acd'] = "acd12_h.dat" #"acd96_h.dat"
    files["H"]['scd'] = "scd12_h.dat" #"scd96_h.dat"
    files["H"]['prb'] = "prb12_h.dat" #"prb96_h.dat"
    files["H"]['plt'] = "plt12_h.dat" #"plt96_h.dat"
    files["H"]['ccd'] = "ccd96_h.dat"
    files["H"]['prc'] = "prc96_h.dat"
    files["H"]['pls'] = "pls_H_14.dat"
    files["H"]['prs'] = "prs_H_14.dat"
    files["H"]['fis'] = "sxrfil14.dat"
    files["H"]['brs'] = "brs05360.dat"
    files["H"]["pbs"] = "pbsx7_h.dat"
    files["H"]["prc"] = "prc89_h.dat"
    files["He"] = {}   #2
    files["He"]['acd'] = "acd96_he.dat"
    files["He"]['scd'] = "scd96_he.dat"
    files["He"]['prb'] = "prb96_he.dat"
    files["He"]['plt'] = "plt96_he.dat"
    files["He"]['ccd'] = "ccd96_he.dat"
    files["He"]['pls'] = "pls_He_14.dat"
    files["He"]['prs'] = "prs_He_14.dat"
    files["He"]['fis'] = "sxrfil14.dat"
    files["He"]['brs'] = "brs05360.dat"
    files["He"]["pbs"] = "pbsx5_he.dat"
    files["He"]['prc'] = "prc96_he.dat"
    files["Li"] = {}   #3
    files["Li"]['acd'] = "acd96_li.dat"
    files["Li"]['scd'] = "scd96_li.dat"
    files["Li"]['ccd'] = "ccd89_li.dat"
    files["Li"]['prb'] = "prb96_li.dat"
    files["Li"]['plt'] = "plt96_li.dat"
    files["Li"]['prc'] = "prc89_li.dat"
    files["Li"]['pls'] = "pls89_li.dat"
    files["Li"]["pbs"] = ''
    files["Li"]["prc"] = "prc89_li.dat"
    files["Be"] = {}   #4
    files["Be"]['acd'] = "acd96_be.dat"
    files["Be"]['scd'] = "scd96_be.dat"
    files["Be"]['prb'] = "prb96_be.dat"
    files["Be"]['plt'] = "plt96_be.dat"
    files["Be"]['ccd'] = "ccd89_be.dat"
    files["Be"]['prc'] = "prc89_be.dat"
    files["Be"]['pls'] = "plsx5_be.dat"
    files["Be"]['prs'] = "prsx5_be.dat"
    files["Be"]["pbs"] = "pbsx5_be.dat"
    files["Be"]["prc"] = "prc89_be.dat"
    files["B"] = {}   #5
    files["B"]['acd'] = "acd74_b.dat"
    files["B"]['scd'] = "scd74_b.dat"
    files["B"]['ccd'] = "ccd89_b.dat"
    files["B"]['prb'] = "prb89_b.dat"
    files["B"]['plt'] = "plt89_b.dat"
    files["B"]['prc'] = "prc89_b.dat"
    files["B"]['pls'] = "plsx5_b.dat"
    files["B"]['prs'] = "prsx5_b.dat"
    files["B"]["pbs"] = "pbsx5_b.dat"
    files["B"]["prc"] = "prc89_b.dat"
    files["C"] = {}    #6
    files["C"]['acd'] = "acd96_c.dat"
    files["C"]['scd'] = "scd96_c.dat"
    files["C"]['prb'] = "prb96_c.dat"
    files["C"]['plt'] = "plt96_c.dat"
    files["C"]['ccd'] = "ccd96_c.dat"
    files["C"]['pls'] = "pls_C_14.dat"
    files["C"]['prs'] = "prs_C_14.dat"
    files["C"]['fis'] = "sxrfil14.dat"
    files["C"]['brs'] = "brs05360.dat"
    files["C"]["pbs"] = "pbsx5_c.dat"
    files["C"]['prc'] = "prc96_c.dat"
    files["N"] = {}    #7
    files["N"]['acd'] = "acd96_n.dat"
    files["N"]['scd'] = "scd96_n.dat"
    files["N"]['ccd'] = "ccd89_n.dat"
    files["N"]['prb'] = "prb96_n.dat"
    files["N"]['plt'] = "plt96_n.dat"
    files["N"]['pls'] = "plsx8_n.dat"
    files["N"]['prs'] = "prsx8_n.dat"
    files["N"]['fis'] = "sxrfilD1.dat"
    files["N"]['brs'] = "brs05360.dat"
    files["N"]['ccd'] = "ccd96_n.dat"
    files["N"]["pbs"] = "pbsx5_n.dat"
    files["N"]["prc"] = "prc96_n.dat"
    files["O"] = {}    #8
    files["O"]['acd'] = "acd96_o.dat"
    files["O"]['scd'] = "scd96_o.dat"
    files["O"]['ccd'] = "ccd89_o.dat"
    files["O"]['prb'] = "prb96_o.dat"
    files["O"]['plt'] = "plt96_o.dat"
    files["O"]['pls'] = "plsx5_o.dat"
    files["O"]['prs'] = "prsx5_o.dat"
    files["O"]["pbs"] = "pbsx5_o.dat"
    files["O"]["prc"] = "prc89_o.dat"
    files["F"] = {}    #9
    files["F"]['acd'] = "acd89_f.dat"  # sub with Puetterich data when available from OPEN-ADAS
    files["F"]['scd'] = "scd89_f.dat"  # sub with Puetterich data when available from OPEN-ADAS
    files["F"]['ccd'] = "ccd89_f.dat"
    files["F"]['prb'] = "prb89_f.dat"
    files["F"]['plt'] = "plt89_f.dat"
    files["F"]['fis'] = "sxrfil14.dat"
    files["F"]['brs'] = "brs05360.dat"
    files["F"]['pls'] = "pls_F_14.dat"
    files["F"]['prs'] = "prs_F_14.dat"
    files["F"]["pbs"] = "pbsx5_f.dat"
    files["F"]['prc'] = "prc89_f.dat"
    files["Ne"] = {}   #10
    files["Ne"]['acd'] = "acd96_ne.dat"
    files["Ne"]['scd'] = "scd96_ne.dat"
    files["Ne"]['prb'] = "prb96_ne.dat"
    files["Ne"]['plt'] = "plt96_ne.dat"
    files["Ne"]['ccd'] = "ccd89_ne.dat"
    files["Ne"]['pls'] = "plsx8_ne.dat"
    files["Ne"]['prs'] = "prsx8_ne.dat"
    files["Ne"]['fis'] = "sxrfilD1.dat"
    files["Ne"]['brs'] = "brs05360.dat"
    files["Ne"]["pbs"] = "pbsx5_ne.dat"
    files["Ne"]['prc'] = "prc89_ne.dat"
    files["Na"] = {}   #10
    files["Na"]['acd'] = "acd85_ne.dat"
    files["Na"]['scd'] = "scd85_ne.dat"
    files["Na"]['prb'] = ""
    files["Na"]['plt'] = ""
    files["Na"]['ccd'] = ""
    files["Na"]['pls'] = ""
    files["Na"]['prs'] = ""
    files["Na"]['fis'] = ""
    files["Na"]['brs'] = ""
    files["Na"]["pbs"] = ""
    files["Na"]['prc'] = ""
    files["Mg"] = {}   #12
    files["Mg"]['acd'] = "acd85_mg.dat"
    files["Mg"]['scd'] = "scd85_mg.dat"
    files["Mg"]['prb'] = ""
    files["Mg"]['plt'] = ""
    files["Mg"]['ccd'] = ""
    files["Mg"]['pls'] = ""
    files["Mg"]['prs'] = ""
    files["Mg"]['fis'] = ""
    files["Mg"]['brs'] = ""
    files["Mg"]["pbs"] = ""
    files["Mg"]['prc'] = ""   
    files["Al"] = {}    #13
    files["Al"]['acd'] = "acd89_al.dat"
    files["Al"]['scd'] = "scd89_al.dat"
    files["Al"]['prb'] = "prb89_al.dat"
    files["Al"]['plt'] = "plt89_al.dat"
    files["Al"]['ccd'] = "ccd89_al.dat"
    files["Al"]['pls'] = "pls_Al_14.dat"
    files["Al"]['prs'] = "prs_Al_14.dat"
    files["Al"]['fis'] = "sxrfil14.dat"
    files["Al"]['brs'] = "brs05360.dat"
    files["Al"]['pbs'] = "pbsx5_al.dat"
    files["Al"]['prc'] = "prc89_al.dat"
    files["Si"] = {}     #14
    files["Si"]['acd'] = "acd96_si.dat"
    files["Si"]['scd'] = "scd96_si.dat"
    files["Si"]['prb'] = "prb96_si.dat"
    files["Si"]['plt'] = "plt96_si.dat"
    files["Si"]['pls'] = "pls_Si_14.dat"
    files["Si"]['prs'] = "prs_Si_14.dat"
    files["Si"]['fis'] = "sxrfil14.dat"
    files["Si"]['brs'] = "brs05360.dat"
    files["Si"]['ccd'] = "ccd89_si.dat"
    files["Si"]["pbs"] = "pbsx5_si.dat"
    files["Si"]["prc"] = "prc89_si.dat"
    files["S"] = {}     #16
    files["S"]['acd'] = "acd89_s.dat"
    files["S"]['scd'] = "scd89_s.dat"
    files["S"]['prb'] = "prb89_s.dat"
    files["S"]['plt'] = "plt89_s.dat"
    files["S"]['pls'] = "pls_S_14.dat"
    files["S"]['prs'] = "prs_S_14.dat"
    files["S"]['fis'] = "sxrfil14.dat"
    files["S"]['brs'] = "brs05360.dat"
    files["S"]['ccd'] = "ccd89_s.dat"
    files["S"]["pbs"] = "pbsx5_s.dat"
    files["S"]["prc"] = "prc89_s.dat"
    files["Cl"] = {}     #17
    files["Cl"]['acd'] = "acd89_cl.dat"
    files["Cl"]['scd'] = "scd89_cl.dat"
    files["Cl"]['prb'] = "prb89_cl.dat"
    files["Cl"]['plt'] = "plt89_cl.dat"
    files["Cl"]['pls'] = "pls_Cl_14.dat"
    files["Cl"]['prs'] = "prs_Cl_14.dat"
    files["Cl"]['fis'] = "sxrfil14.dat"
    files["Cl"]['brs'] = "brs05360.dat"
    files["Cl"]['ccd'] = "ccd89_cl.dat"
    files["Cl"]["pbs"] = "pbsx5_cl.dat"
    files["Cl"]["prc"] = "prc89_cl.dat"
    files["Ar"] = {}     #18
    files["Ar"]['acd'] = "acd89_ar.dat"
    files["Ar"]['scd'] = "scd89_ar.dat"
    files["Ar"]['prb'] = "prb89_ar.dat"
    files["Ar"]['plt'] = "plt41_ar.dat"  # Henderson data
    files["Ar"]['ccd'] = "ccd89_ar.dat"
    files["Ar"]['prc'] = "prc89_ar.dat"
    files["Ar"]['pls'] = "pls_Ar_14.dat"
    files["Ar"]['prs'] = "prs_Ar_14.dat"
    files["Ar"]['fis'] = "sxrfil14.dat"
    files["Ar"]['brs'] = "brs05360.dat"
    files["Ar"]["pbs"] = "pbsx5_ar.dat"
    files["Ar"]["prc"] = "prc89_ar.dat"
    files["Ca"] = {}     #20
    files["Ca"]['acd'] = "acd85_ca.dat"
    files["Ca"]['scd'] = "scd85_ca.dat"
    files["Ca"]['ccd'] = "ccd89_w.dat"  #Ca CCD file not available, use first 20 ion stages using Foster scaling
    files["Ca"]['prb'] = "prb85_ca.dat" # not public on OPEN-ADAS, must request 
    files["Ca"]['plt'] = "plt85_ca.dat" # not public on OPEN-ADAS, must request 
    files["Ca"]['pls'] = "pls_Ca_14.dat"
    files["Ca"]['prs'] = "prs_Ca_14.dat"
    files["Ca"]['fis'] = "sxrfil14.dat"
    files["Ca"]['brs'] = "brs05360.dat"
    files["Ca"]["pbs"] = ""
    files["Ca"]["prc"] = ""
    files["Fe"] = {}     #26
    files["Fe"]['acd'] = "acd89_fe.dat"
    files["Fe"]['scd'] = "scd89_fe.dat"
    files["Fe"]['prb'] = "prb89_fe.dat"
    files["Fe"]['plt'] = "plt41_fe.dat"  # Henderson data
    files["Fe"]['pls'] = "pls_Fe_14.dat"
    files["Fe"]['prs'] = "prs_Fe_14.dat"
    files["Fe"]['fis'] = "sxrfil14.dat"
    files["Fe"]['brs'] = "brs05360.dat"
    files["Fe"]['ccd'] = "ccd89_fe.dat"
    files["Fe"]["pbs"] = "pbsx5_fe.dat"
    files["Fe"]["prc"] = "prc89_fe.dat"
    files["Ni"] = {}     #28
    files["Ni"]['acd'] = "acd85_ni.dat"
    files["Ni"]['scd'] = "scd85_ni.dat"
    files["Ni"]['prb'] = "prb89_ni.dat"
    files["Ni"]['plt'] = "plt89_ni.dat"
    files["Ni"]['pls'] = "pls_Ni_14.dat"
    files["Ni"]['prs'] = "prs_Ni_14.dat"
    files["Ni"]['fis'] = "sxrfil14.dat"
    files["Ni"]['brs'] = "brs05360.dat"
    files["Ni"]['ccd'] = "ccd89_ni.dat"
    files["Ni"]["pbs"] = "pbsx5_ni.dat"
    files["Ni"]["prc"] = "prc89_ni.dat"
    files["Kr"] = {}     #36
    files["Kr"]['acd'] = "acd89_kr.dat"
    files["Kr"]['scd'] = "scd89_kr.dat"
    files["Kr"]['ccd'] = "ccd89_kr.dat"
    files["Kr"]['prb'] = "prb89_kr.dat"
    files["Kr"]['plt'] = "plt41_kr.dat" # Henderson data
    files["Kr"]['pls'] = "plsx5_kr.dat"
    files["Kr"]['prs'] = "prsx5_kr.dat"
    files["Kr"]["pbs"] = "pbsx5_kr.dat"
    files["Kr"]["prc"] = "prc89_kr.dat"
    files["Mo"] = {}     #42
    files["Mo"]['acd'] = "acd89_mo.dat"
    files["Mo"]['scd'] = "scd89_mo.dat"
    files["Mo"]['ccd'] = "ccd89_mo.dat"
    files["Mo"]['plt'] = "plt89_mo.dat"
    files["Mo"]['prb'] = "prb89_mo.dat"
    files["Mo"]['pls'] = "plsx5_mo.dat"
    files["Mo"]['prs'] = "prsx5_mo.dat"
    files["Mo"]["pbs"] = "pbsx5_mo.dat"
    files["Mo"]['prc'] = "prc89_mo.dat"
    files["Xe"] = {}     #56
    files["Xe"]['acd'] = "acd89_xe.dat"
    files["Xe"]['scd'] = "scd89_xe.dat"
    files["Xe"]['ccd'] = "ccd89_xe.dat"
    files["Xe"]['plt'] = "plt41_xe.dat" # Henderson data
    files["Xe"]['prb'] = "prb89_xe.dat"
    files["Xe"]['prs'] = "prsx5_xe.dat"
    files["Xe"]['pls'] = "prsx5_xe.dat"
    files["Xe"]["pbs"] = "pbsx5_xe.dat"
    files["Xe"]['prc'] = "prc89_xe.dat"
    files["W"] = {}     #74
    files["W"]['acd'] = "acd89_w.dat"
    files["W"]['scd'] = "scd89_w.dat"
    files["W"]['prb'] = "prb89_w.dat"
    files["W"]['plt'] = "plt41_w.dat"
    files["W"]['fis'] = "sxrfil14.dat"
    files["W"]['brs'] = "brs05360.dat"
    files["W"]['pls'] = "pls_W_14.dat"
    files["W"]['prs'] = "prs_W_14.dat"
    files["W"]['ccd'] = "ccd89_w.dat"
    files["W"]["pbs"] = "pbsx5_w.dat"
    files["W"]['prc'] = "prc89_w.dat"

    return files



name='aurora'
__version__="0.2.7"

import numpy as np, os

from .core import *
from .atomic import *
from .adas_files import *

from .source_utils import *
from .default_nml import *
from .grids_utils import *
from .coords import *
from .radiation import *

from .particle_conserv import *
from .plot_tools import *
from .animate import *

from .janev_smith_rates import *
from .nbi_neutrals import *
from .neutrals import *

from .synth_diags import *

from .solps import *
from .kn1d import *

aurora_dir = os.path.dirname(os.path.abspath(__file__))



import matplotlib.pyplot as plt
import xarray
import numpy as np
import copy
from scipy.integrate import cumtrapz

def vol_int(Raxis_cm, ds, var, rhop_max=None):
    """
    Perform a volume integral of an input variable. If the variable is f(t,x) 
    then the result is f(t). If the variable is f(t,*,x) then the result is f(t,charge)
    when "*" represents charge, line index, etc...

    Parameters
    ----------
    Raxis_cm : float 
        Major radius on axis [cm]
    ds: xarray dataset 
        Dataset containing Aurora or STRAHL result
    var: str
        Name of the variable in the strahl_result.cdf file
    rhop_max : float
        Maximum normalized poloidal flux for integral. If not provided, integrate
        over the entire simulation grid. 

    Returns
    -------
    var_volint : array (nt,)
        Time history of volume integrated variable
    """
    C = 2 * np.pi * Raxis_cm
    zvol = C * np.pi * ds['rvol_grid'].data / ds['pro'].data 

    # Get our variable
    f = ds[var].data

    # Determine range
    if rhop_max is not None:
        wh = ( ds['rhop_grid'].data <= rhop_max)
        zvol = zvol[wh]
        f  = f[...,wh]
    
    # 2D or 3D array f(t,x)
    var_volint = np.nansum(f*zvol,axis=-1)

    return var_volint




def check_particle_conserv(Raxis_cm, ds=None, filepath=None, linestyle='-', plot=True, axs = None):
    ''' Check time evolution and particle conservation in Aurora or STRAHL output.

    Parameters
    ----------
    Raxis_cm : float
        Major radius on axis [cm], used for volume integrals. 
    ds : xarray dataset, optional
        Dataset containing Aurora results, created using the xarray package. 
        See :py:meth:`~aurora.core.check_conservation` for an illustration on how
        to use this. 
    filepath : str, optional
        If provided, load results from STRAHL output file and check particle
        particle conservation as for an Aurora run. 
    linestyle : str, optional
        matplotlib linestyle, default is '-' (continuous lines). Use this to 
        overplot lines on the same plots using different linestyles, e.g. to check
        whether some aurora option makes particle conservation better or worse. 
    plot : bool, optional
        If True, plot time histories of particle densities in each simulation reservoir.
    axs : 2-tuple or array
        array-like structure containing two matplotlib.Axes instances: the first one 
        for the separate particle time variation in each reservoir, the second for 
        the total particle-conservation check. This can be used to plot results 
        from several aurora runs on the same axes. 

    Returns
    -------
    out : dict
        Dictionary containing time histories across all reservoirs, useful for 
        the assessment of particle conservation.
    axs : 2-tuple or array, only returned if plot=True
        array-like structure containing two matplotlib.Axes instances, (ax1,ax2).
        See optional input argument.
    '''
    if filepath is not None:
        ds = xarray.open_dataset(filepath)  
        # use STRAHL notation for source function
        source_lbl = 'influx_through_valve'
    else:
        # use Aurora notation for source time history
        source_lbl = 'source_time_history'

    # calculate total impurity density (summed over charge states)
    ds['total_impurity_density'] = xarray.DataArray(np.nansum(ds['impurity_density'].data, axis=1),
                                                    coords=[ds['time'].data, ds['rvol_grid'].data], 
                                                    dims=['time', 'space'])
    
    keys = [source_lbl,'particles_in_divertor','particles_in_pump',
            'parallel_loss','parallel_loss_to_limiter','edge_loss', 'particles_at_wall']

    labels = [r'Influx ($s^{-1}$)','Particles in Divertor','Particles in Pump',
              'Parallel Loss',   # losses between LCFS and limiter
              'Parallel Loss to Limiter',  # losses between limiter and edge
              'Edge Loss',  # losses outside the last grid point
              'Particles Stuck at Wall'] # particles that will never leave the wall

    if 'recycling_from_wall' in ds:   # activated recycling model
        if 'particles_retained_at_wall' in ds:    # activated wall retention model
            keys += ['particles_retained_at_wall']
            labels += [ 'Particles Retained at Wall']

        keys += ['recycling_from_wall','recycling_from_divertor']
        labels += ['Wall Recy. Rate','Divertor Recy. Rate']

    for key in keys:
        # substitute 0 to nan's for particle numbers
        ds[key] = copy.deepcopy(ds[key].fillna(0.0))

    time = ds['time'].data
    vol_int_keys = ['total_impurity_density','impurity_radiation']
    vol_int_rhop = [1.0, 1.0]
    vol_int_labels = ['Core Impurity Particles', 'Core Radiation (W)']

    # factor to account for cylindrical geometry:
    circ = 2*np.pi*Raxis_cm # cm

    # Compute total number of particles for particle conservation checks:
    all_particles = vol_int(Raxis_cm,ds, 'total_impurity_density')
    total = all_particles+(ds['particles_at_wall'].data+\
                           ds['particles_in_divertor'].data)*circ 
    
    if 'recycling_from_wall' in ds:
        total+=ds['particles_in_pump'].data*circ
        
        if 'particles_retained_at_wall' in ds:
            total+=ds['particles_retained_at_wall'].data*circ

    # particles that entered as "source":
    source = ds[source_lbl].data*circ
    integ_source =  cumtrapz(source,time,initial=0)
    
    # collect all the relevant quantities for particle conservation
    out = {}
    out['total'] = total
    out['plasma_particles'] = vol_int(Raxis_cm, ds, 'total_impurity_density')
    out['particles_at_wall'] = ds['particles_at_wall'].data*circ,
    out['particles_in_divertor'] = ds['particles_in_divertor'].data*circ
    if 'recycling_from_wall' in ds:
        out['particles_in_pump'] = ds['particles_in_pump'].data*circ
    if 'recycling_from_wall' in ds and 'particles_retained_at_wall' in ds:
        out['particles_retained_at_wall'] = ds['particles_retained_at_wall'].data*circ
    out['integ_source'] = integ_source


    if plot:
        # ---------------------------------------------------------------------
        # plot time histories for each particle reservoirs:
        nplots = np.sum([k in ds for k in keys+vol_int_keys])
        ncol = min(3,nplots)
        nrow = int(np.ceil(float(nplots)/ncol))

        if axs is None:
            fig,ax1 = plt.subplots(nrows=nrow,ncols=ncol,sharex=True, figsize=(15,10))
        else:
            ax1 = axs[0]

        axf = ax1.flatten()
        iplot = 0

        for key,label in zip(keys,labels):
            if key in ds:
                y = ds[key].data
                if key != 'particles_in_pump':
                    y = y* circ #normalize it to #/s or # units

                axf[iplot].plot(time,y,'-o',label=label,markersize=2, ls=linestyle)
                iplot += 1

        for i,key,rhop,lab in zip(list(range(len(vol_int_keys))),vol_int_keys,vol_int_rhop,vol_int_labels):
            if key in ds:
                vol_int_data = vol_int(Raxis_cm, ds, key,rhop_max=rhop)

                if key == 'impurity_radiation':
                    vol_int_data = vol_int_data[:,-1]   #total radiation
                axf[iplot].plot(time,vol_int_data,'o-',label=lab,markersize=2.0, ls=linestyle)
                iplot += 1

        for axxx in axf[-ncol:]:
            axxx.set_xlabel('Time (s)')
        axxx.set_xlim(time[[0,-1]])

        for aaa in axf[:iplot]:
            aaa.legend(loc='best').set_draggable(True)

        # --------------------------------------------------------------------------------------------
        # now plot all particle reservoirs to check particle conservation:

        if axs is None:
            fig,ax2 = plt.subplots()
        else:
            ax2 = axs[1]

        ax2.set_xlabel('time [s]')

        ax2.plot(time, all_particles ,label='Particles in Plasma', ls=linestyle)
        ax2.plot(time, ds['particles_at_wall'].data*circ,label='Particles stuck at wall', ls=linestyle)
        ax2.plot(time, ds['particles_in_divertor'].data*circ,label='Particles in Divertor', ls=linestyle)

        if 'recycling_from_wall' in ds:
            ax2.plot(time, ds['particles_in_pump'].data*circ ,label='Particles in Pump', ls=linestyle)

            if 'particles_retained_at_wall' in ds:
                # only plot particles retained at wall if retention model is used
                ax2.plot(time, ds['particles_retained_at_wall'].data*circ,label='Particles retained at wall', ls=linestyle)

        ax2.plot(time,total,lw=2,label='Total', ls=linestyle)

        ax2.plot(time,integ_source,lw=2,label='Integrated source', ls=linestyle)

        if abs((total[-1]-integ_source[-1])/integ_source[-1])> .1:
            print('Warning: significant error in particle conservation!')   #try increasing grid or time resolution

        ax2.set_ylim(0,None)
        ax2.legend(loc='best')

        plt.tight_layout()

    
    # close dataset if necessary
    try:
        ds.close()
    except:
        pass

    if plot:
        return out, (ax1,ax2)
    else:
        return out




'''Methods to create radial and time grids for aurora simulations.
'''

import matplotlib.pyplot as plt
import numpy as np, sys, os
from scipy.interpolate import interp1d
import warnings

def create_radial_grid(namelist,plot=False):
    r'''Create radial grid for Aurora based on K, dr_0, dr_1, rvol_lcfs and bound_sep parameters. 
    The lim_sep parameters is additionally used if plotting is requested. 

    Radial mesh points are set to be equidistant in the coordinate :math:`\rho`, with

    .. math::

        \rho = \frac{r}{\Delta r_{centre}} + \frac{r_{edge}}{k+1} \left(\frac{1}{\Delta r_{edge}}- \frac{1}{\Delta r_{centre}} \right) \left(\frac{r}{r_{edge}} \right)^{k+1}

    The corresponding radial step size is 

    .. math::

        \Delta r = \left[\frac{1}{\Delta r_{centre}} + \left(\frac{1}{\Delta r_{edge}} - \frac{1}{\Delta r_{centre}} \right) \left(\frac{r}{r_{edge}}\right)^k \right]^{-1}


    See the STRAHL manual for details. 

    Parameters
    ----------
    namelist : dict
        Dictionary containing Aurora namelist. This function uses the K, dr_0, dr_1, rvol_lcfs 
        and bound_sep parameters. Additionally, lim_sep is used if plotting is requested. 
    plot : bool, optional
        If True, plot the radial grid spacing vs. radial location. 

    Returns
    -------
    rvol_grid : array
        Volume-normalized grid used for Aurora simulations.
    pro : array
        Normalized first derivatives of the radial grid, defined as 
        pro = (drho/dr)/(2 d_rho) = rho'/(2 d_rho)
    qpr : array
        Normalized second derivatives of the radial grid, defined as 
        qpr = (d^2 rho/dr^2)/(2 d_rho) = rho''/(2 d_rho)
    prox_param : float
        Grid parameter used for perpendicular loss rate at the last radial grid point. 
    '''
    
    K = namelist['K']
    dr_0 = namelist['dr_0']
    dr_1 = namelist['dr_1']
    rvol_lcfs = namelist['rvol_lcfs']
    dbound = namelist['bound_sep']

    # radial location of wall boundary:
    r_edge = namelist['rvol_lcfs'] + namelist['bound_sep']

    try:
        assert dr_0>0.0
        assert dr_1>0.0
    except:
        raise ValueError('Both dr_0 and dr_1 must be positive!')
    
    a0=1./dr_0

    # number of radial points:
    ir=int(round(1.5+r_edge*(a0*K+ (1./dr_1))/(K+1.)))

    rvol_grid = np.zeros(ir)
    pro_grid = np.zeros(ir)
    qpr_grid = np.zeros(ir)

    # prox_param is used in calculation of ion perpendicular loss at the last grid point
    prox_param=(K+1.)*(float(ir)-1.5)/r_edge-a0*K

    # form the radial grid iteratively
    rvol_grid[0] = 0.
    for i in np.arange(1,ir):
        temp1 = 0.
        temp2 = r_edge*1.05
        
        for j in np.arange(50):
            rvol_grid[i] = (temp1+temp2)/2.
            temp3 = a0*rvol_grid[i]+(prox_param-a0)*r_edge/(K+1.)*(rvol_grid[i]/r_edge)**(K+1.)
            if temp3 > i:
                temp2 = rvol_grid[i]
            else:
                temp1 = rvol_grid[i]

    # terms related with derivatives of rho
    temp1 = 0.5
    pro_grid[0] = 2./(dr_0**2)
    for i in np.arange(1,ir):
        pro_grid[i] = (a0+(prox_param-a0)*(rvol_grid[i]/r_edge)**K)*temp1
        qpr_grid[i] = pro_grid[i]/rvol_grid[i]+temp1*(prox_param-a0)*K/r_edge*(rvol_grid[i]/r_edge)**(K-1.)

    if plot:

        r_lim  = namelist['rvol_lcfs'] + namelist['lim_sep']
        dr = np.gradient(rvol_grid)

        if plt.fignum_exists('Aurora radial step'):
            plt.figure(num='Aurora radial step').clf()
        f,ax = plt.subplots(num='Aurora radial step')
        
        ax.plot(rvol_grid/namelist['rvol_lcfs'], dr,'-')
        ax.axvline(1,ls='--',c='k')
        ax.text(1+0.01,namelist['dr_0']*0.9 ,'LCFS',rotation='vertical')

        ax.axvline(r_lim/namelist['rvol_lcfs'],ls='--',c='k')
        ax.text(r_lim/namelist['rvol_lcfs']+0.01,namelist['dr_0']*0.9 ,'limiter',rotation='vertical')

        if 'saw_model' in namelist and namelist['saw_model']['saw_flag']:
            ax.axvline( namelist['saw_model']['rmix']/namelist['rvol_lcfs'],ls='--',c='k')
            ax.text(namelist['saw_model']['rmix']/namelist['rvol_lcfs']+0.01,
                    namelist['dr_0']*0.5 ,'Sawtooth mixing radius',rotation='vertical')

        ax.set_xlabel(r'$r/r_{lcfs}$');
        ax.set_ylabel(r'$\Delta$ r [cm]');
        ax.set_ylim(0,None)
        ax.set_xlim(0,r_edge/namelist['rvol_lcfs'])
        ax.set_title(r'$\#$ radial grid points: %d'%len(rvol_grid))

    return rvol_grid, pro_grid, qpr_grid, prox_param




class MissingAuroraBuild(UserWarning):
    pass
    
def create_time_grid(timing=None, plot=False):
    '''Create time grid for simulations using the Fortran implementation
    of the time grid generator. 

    Parameters
    ----------
    timing : dict
        Dictionary containing timing elements: 'times', 'dt_start', 'steps_per_cycle','dt_increase'
        As in STRAHL, the last element in each of these arrays refers to sawtooth events. 
    plot : bool
        If True, plot time grid. 

    Returns
    -------
    time : array
        Computational time grid corresponding to :param:timing input.
    save : array
        Array of zeros and ones, where ones indicate that the time step will be stored in memory
        in Aurora simulations. Points corresponding to zeros will not be returned to spare memory. 
    '''
    # import here to avoid import when building documentation or package
    try:
        from ._aurora import time_steps
    except ModuleNotFoundError:
        raise MissingAuroraBuild('Could not load particle transport forward model!'+\
                      'Use the makefile or setup.py to build sources.')
    
    _time, _save = time_steps(
        timing['times'],timing['dt_start'],timing['steps_per_cycle'],timing['dt_increase'])

    # eliminate trailing 0's:
    idxs = np.nonzero(_time)
    time = _time[idxs]
    save = _save[idxs] > 0
    
    if plot:
        #show timebase
        if plt.fignum_exists('Aurora time step'):
            plt.figure('Aurora time step').clf()
        f,ax = plt.subplots(num='Aurora time step')
        ax.set_title(r'$\#$ time steps: %d   $\#$ saved steps %d'%(len(time), sum(save)))
        ax.semilogy(time[1:],np.diff(time),'.-',label='step')
        ax.semilogy(time[1:][save[1:]],np.diff(time)[save[1:]],'o',label='step')
        [ax.axvline(t,c='k',ls='--') for t in timing['times']]
        ax.set_xlim(time[0],time[-1])

    return time, save



    
def create_time_grid_new(timing, verbose=False, plot=False):
    '''Define time base for Aurora based on user inputs
    This function reproduces the functionality of STRAHL's time_steps.f
    Refer to the STRAHL manual for definitions of the time grid
    
    Parameters
    ----------
    n : int
        Number of elements in time definition arrays
    t : array
        Time vector of the time base changes
    dtstart : array
        dt value at the start of a cycle
    itz : array
        cycle length, i.e. number of time steps before increasing dt
    tinc :
        factor by which time steps should be increasing within a cycle
    verbose : bool
        If Trueprint to terminal a few extra info
    
    Returns
    -------
    t_vals : array
        Times in the time base [s]
    i_save : array
        Array of 0,1 values indicating at which times internal arrays should be stored/returned. 

    
    ~~~~~~~~~~~ THIS ISN'T FUNCTIONAL YET! ~~~~~~~~~~~~


    '''
    t = np.array(timing['times'])
    dtstart = np.array(timing['dt_start'])
    itz = np.array(timing['steps_per_cycle'])
    tinc = np.array(timing['dt_increase'])
    
    t_vals = np.zeros(30000)
    i_save = np.zeros(30000)

    dt = np.zeros(250)
    ncyctot = np.zeros(250)    
    ncyc = np.zeros(250)
    
    # zeross without double time points
    t_s = np.zeros(250)
    dtstart_s = np.zeros(250)
    tinc_s = np.zeros(250)
    itz_s = np.zeros(250)
    
    # sort t-change and related verctors
    idxs = np.argsort(t)
    dtstart = dtstart[idxs]
    tinc = tinc[idxs]
    itz = itz[idxs]
    
    if verbose:
        print('Inputs after sorting:')
        print('t:',t)
        print('dtstart: ', dtstart)
        print('tinc: ', tinc)
        print('itz: ', itz)
        
    # cancel double time points
    nevent = 1
    t_s[0] = t[0]
    dtstart_s[0] = dtstart[0]
    tinc_s[0] = tinc[0]
    itz_s[0] = itz[0]
    for i in np.arange(1,len(t)):
        if abs(t[i]-t[nevent]) > 1e-8:
            nevent = nevent + 1
            t_s[nevent] = t[i]
            dtstart_s[nevent] = dtstart[i]
            tinc_s[nevent] = tinc[i]
            itz_s[nevent] = itz[i]

    # define # of cycles for every interval with a start time step: dtstart[i]
    for i in np.arange(nevent-1):
        f = (t_s[i+1] - t_s[i])/dtstart_s[i]
        if i==0: f = f + 1.
        if tinc_s[i]>1.:
            ncyc[i] = max(2,int(np.log(f/itz_s[i]*(tinc_s[i]-1.)+1.)/np.log(tinc_s[i])))
            
        if (tinc_s[i]==1.):
            ncyc[i] = max(2,int(f/itz_s[i]))

        if i==0: ncyctot[i] = ncyc[i]
        if i>0: ncyctot[i] = ncyctot[i-1] + ncyc[i]

    # sum of all timesteps
    nsteps = 0
    for i in np.arange(nevent-1):
        nsteps = nsteps + ncyc[i] * itz_s[i]
        
    nsteps = nsteps - 1

    # define real start timestep dt[i] to fit the time intervals
    for i in np.arange(nevent-1):
        if tinc_s[i]>1.:
            f = itz_s[i]* (tinc_s[i]**ncyc[i]-1.)/(tinc_s[i]-1.)

        if tinc_s[i]==1.: f = 1. * itz_s[i] * ncyc[i]
        if i==1: f = f - 1.
        dt[i] = (t_s[i+1] - t_s[i])/f

    # Now, define t list
    nn = 1
    n_itz = 1
    m = 1
    tnew = t_s[0]
    nevent=1
    det = dt[0]

    if verbose:
        print('time_steps:')
        print('nsteps:',nsteps)
        for i in np.arange(nevent-1):
            print('Interval:',i,' start step: ',dt[i])
    
    t_vals[0] = tnew
    i_save[0] = 1

    tmp2 = True
    while tmp2:
        tmp = True
        while tmp:
            tnew = tnew + det
            t_vals[nn+1] = tnew

            if np.mod(n_itz+1,itz_s[nevent])!= 0:
                nn = nn+1
                n_itz = n_itz + 1
            else:
                tmp=False

        n_itz = 0
        det = tinc[nevent] * det
        if (m == ncyctot[nevent]) & (nn != nsteps):
            nevent = nevent + 1
            det = dt[nevent]

        m=m+1
        i_save[nn+1] = 1

        if nn<nsteps:
            nn = nn + 1
        else:
            tmp2=False


    # -------------
    # eliminate trailing 0's:
    idxs = np.nonzero(t_vals)
    time = t_vals[idxs]
    save = i_save[idxs] > 0

    if plot:
        #show timebase
        if plt.fignum_exists('Aurora time step'):
            plt.figure('Aurora time step').clf()
        f,ax = plt.subplots(num='Aurora time step')
        ax.set_title('# time steps: %d    # saved steps %d'%(len(time), sum(save)))
        ax.semilogy(time[1:],np.diff(time),'.-',label='step')
        ax.semilogy(time[1:][save[1:]],np.diff(time)[save[1:]],'o',label='step')
        [ax.axvline(t,c='k',ls='--') for t in timing['times']]
        ax.set_xlim(time[0],time[-1])

    return time, save



def get_HFS_LFS(geqdsk, rho_pol=None):
    '''Get high-field-side (HFS) and low-field-side (LFS) major radii from the g-EQDSK data. 
    This is useful to define the rvol grid outside of the LCFS. 
    See the :py:func:`~aurora.grids_utils.get_rhopol_rvol_mapping` for an application. 

    Parameters
    ----------
    geqdsk : dict
        Dictionary containing the g-EQDSK file as processed by the `omfit_classes.omfit_eqdsk`. 
    rho_pol : array, optional
        Array corresponding to a grid in sqrt of normalized poloidal flux for which a 
        corresponding rvol grid should be found. If left to None, an arbitrary grid will be 
        created internally. 
    
    Returns
    -------
    Rhfs : array
        Major radius [m] on the HFS
    Rlfs : array
        Major radius [m] on the LFS

    '''
    if rho_pol is None:
        rho_pol = np.linspace(0,1.2,70)

    R = geqdsk['AuxQuantities']['R']
    Z = geqdsk['AuxQuantities']['Z']

    Z0 = geqdsk['fluxSurfaces']['Z0']
    R0 = geqdsk['fluxSurfaces']['R0']
    RHORZ = geqdsk['AuxQuantities']['RHOpRZ']

    rho_mid = interp1d(Z,RHORZ,axis=0)(Z0)
    center = R.searchsorted(R0)
    rho_mid[:center] *= -1

    #remove a small discontinuity in gradients on axis
    R = np.delete(R,[center-1,center])
    rho_mid = np.delete(rho_mid,[center-1,center])

    Rhfs = np.interp(rho_pol, -rho_mid[::-1],R[::-1])
    Rlfs = np.interp(rho_pol, rho_mid,R)

    return Rhfs, Rlfs



def get_rhopol_rvol_mapping(geqdsk, rho_pol=None):
    r'''Compute arrays allowing 1-to-1 mapping of rho_pol and rvol, both inside and
    outside the LCFS.

    rvol is defined as :math:`\sqrt{V/(2 \pi^2 R_{axis}}` inside the LCFS. Outside of it,
    we artificially expand the LCFS to fit true equilibrium at the midplane based
    on the rho_pol grid (sqrt of normalized poloidal flux).

    Method:
    
    .. math::

        r(\rho,\theta) = r_0(\rho) +  (r_{lcfs}(\theta) - r_{0,lcfs}) \times \mathcal{f} \\
        z(\rho,\theta) = z_0      +  (z_{lcfs}(\theta) - z_0     ) \times \mathcal{f} \\
        \mathcal{f} = \frac{ r(\rho,\theta=0) - r(\rho,\theta=180)}{r_{lcfs}(\theta=0)- r_{lcfs}(\theta=180)} \\
        r_{0,lcfs} = \frac{1}{2} (r_{lcfs}(\theta=0)+ r_{lcfs}(\theta=180)) \\
        r_0(\rho) = \frac{1}{2} (r(\rho,\theta=0) + r(\rho,\theta=180))

    The mapping between rho_pol and rvol allows one to interpolate inputs on a
    rho_pol grid onto the rvol grid (in cm) used internally by the code.

    Parameters
    ----------
    geqdsk : dict
        Dictionary containing the g-EQDSK file as processed by `omfit_classes.omfit_eqdsk`. 
    rho_pol : array, optional
        Array corresponding to a grid in sqrt of normalized poloidal flux for which a 
        corresponding rvol grid should be found. If left to None, an arbitrary grid will be 
        created internally. 

    Returns
    -------
    rho_pol : array
        Sqrt of normalized poloidal flux grid
    rvol : array
        Mapping of rho_pol to a radial grid defined in terms of normalized flux surface volume.
    '''

    if rho_pol is None:
        # use arbitrary rho_pol grid, equally-spaced
        rho_pol = np.linspace(0.0, 1.1, 79)

    # volumes calculated by EFIT
    R0 = geqdsk['RMAXIS']
    Z0 = geqdsk['ZMAXIS']
    
    V_inner = geqdsk['fluxSurfaces']['geo']['vol']
    rhop_inner = np.sqrt( geqdsk['fluxSurfaces']['geo']['psin'])

    # find Rlfs and Rhfs for each value of rho_pol
    Rhfs, Rlfs = get_HFS_LFS(geqdsk, rho_pol)

    # R and Z along the LCFS
    R_lcfs = geqdsk['RBBBS']
    Z_lcfs = geqdsk['ZBBBS']

    r0_lcfs = 0.5*(np.interp(1,rho_pol,Rhfs) + np.interp(1,rho_pol,Rlfs))
    r0 = 0.5*(Rhfs + Rlfs)

    scale = (Rlfs-Rhfs)/(np.interp(1,rho_pol,Rlfs) - np.interp(1,rho_pol,Rhfs))
    R_outer = r0 + np.outer(R_lcfs - r0_lcfs,scale)
    Z_outer = Z0 + np.outer(Z_lcfs - Z0     ,scale)

    # calculate volume enclosed by these "flux surfaces" outside of the LCFS
    #V_outer = -sum(2*np.pi*((R_outer+np.roll(R_outer,1,0))/2.)**2*(Z_outer-np.roll(Z_outer,1,0)),0)/2.0
    V_outer = np.abs(sum(2*np.pi*((R_outer+np.roll(R_outer,1,0))/2.)**2*(Z_outer-np.roll(Z_outer,1,0)),0)/2.0)
    
    V = np.interp(rho_pol,rhop_inner, V_inner)
    V[rho_pol > 1] = V_outer[rho_pol > 1]

    # correct errors in V close to magnetic axis inside of rho = .2
    V[rho_pol <.2] = V_outer[rho_pol <.2]/V_outer[rho_pol <.2][-1]*V[rho_pol <.2][-1]

    # compute rvol
    rvol = np.sqrt(V/(2*np.pi**2 * R0)) * 100 # m --> cm

    # enforce 0 on axis
    rvol[0] = 0.0
    
    return rho_pol, rvol






def create_radial_grid_fortran(namelist,plot=False):
    '''This interfaces the package subroutine to create the radial grid exactly as STRAHL does it.
    Refer to the STRAHL manual for details.

    '''
    # import here to avoid import when building documentation or package
    try:
        from ._aurora import get_radial_grid
    except ModuleNotFoundError:
        raise MissingAuroraBuild('Could not load particle transport forward model!'+\
                      'Use the makefile or setup.py to build sources.')
    
    # NB: there is currently a hard-coded maximum number of grid points (1000)
    _r, _pro, prox, _qpr = get_radial_grid(
        namelist['ng'],namelist['bound_sep'],namelist['K'],namelist['dr_0'],namelist['dr_1'], namelist['rvol_lcfs'])

    # eliminate trailing zeros:
    idxs = _r > 0
    idxs[0] = True
    radius_grid, pro, qpr = _r[idxs], _pro[idxs], _qpr[idxs]

    if plot:

        r_lim  = namelist['rvol_lcfs'] + namelist['lim_sep']
        r_wall = namelist['rvol_lcfs'] + namelist['bound_sep']
        dr = np.gradient(radius_grid)

        if plt.fignum_exists('aurora radial step'):
            plt.figure(num='aurora radial step').clf()
        f,ax = plt.subplots(num='aurora radial step')
        
        ax.plot(radius_grid/namelist['rvol_lcfs'], dr,'-')
        ax.axvline(1,ls='--',c='k')
        ax.text(1+0.01,namelist['dr_0']*0.9 ,'LCFS',rotation='vertical')

        ax.axvline(r_lim/namelist['rvol_lcfs'],ls='--',c='k')
        ax.text(r_lim/namelist['rvol_lcfs']+0.01,namelist['dr_0']*0.9 ,'limiter',rotation='vertical')

        if 'saw_model' in namelist and namelist['saw_model']['saw_flag']:
            ax.axvline( namelist['saw_model']['rmix']/namelist['rvol_lcfs'],ls='--',c='k')
            ax.text(namelist['saw_model']['rmix']/namelist['rvol_lcfs']+0.01,namelist['dr_0']*0.5 ,
                    'Sawtooth mixing radius',rotation='vertical')

        ax.set_xlabel(r'$r/r_{lcfs}$');
        ax.set_ylabel(r'$\Delta$ r [cm]');
        ax.set_ylim(0,None)
        ax.set_xlim(0,r_wall/namelist['rvol_lcfs'])
        ax.set_title('# radial grid points: %d'%len(radius_grid))

    return radius_grid, pro, prox, qpr



def create_aurora_time_grid(timing, plot=False):
    '''Create time grid for simulations using a Fortran routine for definitions. 
    The same functionality is offered by :py:func:`~aurora.grids_utils.create_time_grid`, which however
    is written in Python. This method is legacy code; it is recommended to use the other. 

    Parameters
    ----------
    timing : dict
        Dictionary containing 
        timing['times'],timing['dt_start'],timing['steps_per_cycle'],timing['dt_increase']
        which define the start times to change dt values at, the dt values to start with,
        the number of time steps before increasing the dt by dt_increase. 
        The last value in each of these arrays is used for sawteeth, whenever these are
        modelled, or else are ignored. This is the same time grid definition as used in STRAHL.
    plot : bool, optional
        If True, display the created time grid.

    Returns
    -------
    time : array
        Computational time grid corresponding to `timing` input.
    save : array
        Array of zeros and ones, where ones indicate that the time step will be stored in memory
        in aurora simulations. Points corresponding to zeros will not be returned to spare memory.    
    '''
    # import here to avoid import when building documentation or package
    from ._aurora import time_steps
    
    _time, _save = _aurora.time_steps(
        timing['times'],timing['dt_start'],timing['steps_per_cycle'],timing['dt_increase'])

    # eliminate trailing 0's:
    idxs = np.nonzero(_time)
    time = _time[idxs]
    save = _save[idxs] > 0

    if plot:
        #show timebase
        if plt.fignum_exists('STRAHL time step'):
            plt.figure('STRAHL time step').clf()
        f,ax = plt.subplots(num='STRAHL time step')
        ax.set_title('# time steps: %d    # saved steps %d'%(len(time), sum(save)))
        ax.semilogy(time[1:],np.diff(time),'.-',label='step')
        ax.semilogy(time[1:][save[1:]],np.diff(time)[save[1:]],'o',label='step')
        [ax.axvline(t,c='k',ls='--') for t in timing['times']]
        ax.set_xlim(time[0],time[-1])

    return time, save




def estimate_clen(geqdsk):
    '''Estimate average connection length in the open SOL and in the limiter shadow
    NB: these are just rough numbers!

    Parameters
    ----------
    geqdsk : dict
        EFIT g-EQDSK as processed by `omfit_classes.omfit_eqdsk`.

    Returns
    -------
    clen_divertor : float
        Estimate of the connection length to the divertor [m]
    clen_limiter : float
        Estimate of the connection length to the limiter [m]
    '''
    # estimate connection legnth in divertor
    q = geqdsk['QPSI']
    rhop = np.sqrt(geqdsk['fluxSurfaces']['geo']['psin'])

    # q at the 95% surface (in sqrtpsinorm)
    q95 = np.abs(interp1d(rhop,q)(0.95))
    R0 = geqdsk['fluxSurfaces']['R0']

    # estimate for connection length to the divertor
    clen_divertor = round(np.pi*R0*q95,5)

    # estimate for connection length to limiter
    zlim = geqdsk['ZLIM']
    h = np.max(zlim) - np.min(zlim)
    clen_limiter = round(h/5.,5) # 1/5th of machine height
    
    return clen_divertor, clen_limiter




def estimate_boundary_distance(shot, device, time_ms):
    '''Obtain a simple estimate for the distance between the LCFS and the wall boundary.
    This requires access to the A_EQDSK on the EFIT01 tree on MDS+. Users who may find that this call
    does not work for their device may try to adapt the OMFITmdsValue TDI string.

    Parameters
    ----------
    shot : int
        Discharge/experiment number
    device : str
        Name of device, e.g. 'C-Mod', 'DIII-D', etc.
    time_ms : int or float
        Time at which results for the outer gap should be taken.

    Returns
    -------
    bound_sep : float
        Estimate for the distance between the wall boundary and the separatrix [cm]
    lim_sep : float
        Estimate for the distance between the limiter and the separatrix [cm]. This is (quite arbitrarily)
        taken to be 2/3 of the bound_sep distance.    
    '''
    # import this here, so that it is not required for the whole package
    from omfit_classes.omfit_mds import OMFITmdsValue

    try:
        tmp = OMFITmdsValue(server=device, treename='EFIT01', shot=shot,
                            TDI='\\EFIT01::TOP.RESULTS.A_EQDSK.ORIGHT')  # CMOD format, take ORIGHT
    except Exception:
        tmp = OMFITmdsValue(server=device, treename='EFIT01', shot=shot,
                            TDI='\\EFIT01::TOP.RESULTS.AEQDSK.GAPOUT') # useful variable for many other devices

    time_vec = tmp.dim_of(0)
    data_r = tmp.data()
    
    ind = np.argmin(np.abs(time_vec - time_ms))
    inds = slice(ind-3, ind+3)
    bound_sep = round(np.mean(data_r[inds]),3)

    # take separation to limiter to be 2/3 of the separation to the wall boundary
    lim_sep = round(bound_sep*2./3.,3)

    return bound_sep, lim_sep



'''Aurora functionality for edge neutral modeling. 
The ehr5 file from DEGAS2 is used. See https://w3.pppl.gov/degas2/ for details.
'''

from matplotlib import cm
from scipy.optimize import curve_fit, least_squares
import numpy as np
import matplotlib.pyplot as plt
plt.ion()
import urllib
import shutil, os, copy
from scipy.constants import e,h,c as c_light,Rydberg
from scipy.interpolate import interp2d

from . import plot_tools
from . import radiation

def download_ehr5_file():
    '''Download the ehr5.dat file containing atomic data describing the multi-step ionization and 
    recombination of hydrogen. 

    See https://w3.pppl.gov/degas2/ for details. 
    '''
    filename='ehr5.dat'
    url = 'https://w3.pppl.gov/degas2/ehr5.dat' 
    local_filename,headers = urllib.request.urlretrieve(url, filename)
    shutil.move(filename, os.path.dirname(os.path.realpath(__file__))+'/ehr5.dat')
    print('Successfully downloaded the DEGAS2 ehr5.dat file.')
    

class ehr5_file:
    ''' Read ehr5.dat file from DEGAS2. 
    Returns a dictionary containing

    - Ionization rate Seff in :math:`cm^3 s^{-1}`
    - Recombination rate Reff in :math:`cm^3 s^{-1}`
    - Neutral electron losses :math:`E_{loss}^{(i)}` in :math:`erg s^{-1}`
    - Continuum electron losses :math:`E_{loss}^{(ii)}` in :math:`erg s^{-1}`
    - Neutral “n=2 / n=1”, :math:`N_2^{(i)}/N_1`
    - Continuum “n=2 / n=1”, :math:`N_2^{(ii)}/N_11
    - Neutral “n=3 / n=1”, :math:`N_3^{(i)}/N_1`
    - Continuum “n=3 / n=1”, :math:`N_3^{(ii)}/N_1`
    
    ... and similarly for n=4 to 9. Refer to the DEGAS2 manual for details.
    '''

    def __init__(self, filepath=None):
        '''Load ehr5.dat file, either from the indicated path or by downloading it locally. 

        Keyword Args:
            filepath : str, optional
                Path of ehr5.dat file to use. If left to None, the file is downloaded from the web
                and saved locally. 

        Results for each of the fields in the `fields` attribute will be available in the `res`
        attribute in the form of a dictionary. Refer to the DEGAS2 manual for a description of these
        fields.
        '''

        if filepath is None:
            if not os.path.exists(os.path.dirname(os.path.realpath(__file__))+'/ehr5.dat'):
                # if ehr5.dat file is not available, download it
                download_ehr5_file()
            self.filepath = os.path.dirname(os.path.realpath(__file__))+'/ehr5.dat'
        else:
            self.filepath = filepath

        self.ne = 10 ** np.array([10 + (jn - 1.0) / 2.0 for jn in np.arange(1, 16)])  # cm^{-3}
        self.Te = 10 ** np.array([-1.2 + (jt - 1.0) / 10.0 for jt in np.arange(1, 61)])  # eV

        self.fields = [
            'Seff',
            'Reff',
            'Ei_loss',
            'Eii_loss',
            'n3i_n1',
            'n3ii_n1',
            'n2i_n1',
            'n2ii_n1',
            'n4i_n1',
            'n4ii_n1',
            'n5i_n1',
            'n5ii_n1',
            'n6i_n1',
            'n6ii_n1',
            'n7i_n1',
            'n7ii_n1',
            'n8i_n1',
            'n8ii_n1',
            'n9i_n1',
            'n9ii_n1',
        ]

        # get data
        self.load()

    def load(self):

        self.res = {}
        with open(self.filepath) as f:
            for field in self.fields:

                data = np.zeros((15, 60))

                # read header
                header = f.readline()

                for jn in np.arange(15):
                    # loop over 15 densities
                    _jn_index = f.readline()
                    arr = []
                    for jt_row in np.arange(10):
                        # 10 rows of 6 values each for Te
                        elems = [val for val in f.readline().strip().split(' ') if val != '']

                        line = [float(val) for val in elems]
                        data[jn, jt_row * 6 : (jt_row + 1) * 6] = np.array(line)
                    _dum = f.readline()  # empty line at the end

                self.res[field] = copy.deepcopy(data)

    def plot(self, field='Seff', fig=None, axes=None):
        colormap = cm.rainbow
        if fig is None or axes is None:
            fig, ax = plt.subplots()

        ls_cycle = plot_tools.get_ls_cycle()

        labels = ['{:.2e}'.format(val) + ' cm${-3}$' for val in self.ne]
        for i in np.arange(len(labels)):
            ax.plot(np.log10(self.Te), self.res[field][i, :], next(ls_cycle), label=labels[i])

        ax.set_xlabel('$\log T_e\ \mathrm{[eV]}$', fontsize=16)
        ax.set_ylabel(field, fontsize=16)
        ax.legend()



def get_exc_state_ratio(m, N1, ni, ne, Te, rad_prof=None, rad_label=r'rmin [cm]', plot=False):
    r"""Compute density of excited states in state `m` (m>1), given the density of ground state atoms.
    This function is not l-resolved.

    The function returns

    .. math::

        N_m/N_1 = \left(\frac{N_m^i}{N_1} \right) N_m + \left(\frac{N_m^{ii}}{n_i} \right) n_i


    where :math:`N_m` is the number of electrons in the excited state :math:`m`, :math:`N_1` 
    is the number in the ground state, and :math:`n_i` is the density of ions that could recombine.
    :math:`i` and :math:`ii` indicate terms corresponding to coupling to the ground state and to 
    the continuum, respectively.

    Ref.: DEGAS2 manual.

    Parameters
    ----------
    m : int
        Principal quantum number of excited state of interest. 2<m<10
    N1 : float, list or 1D-array [:math:`cm^{-3}`]
        Density of ions in the ground state. This must have the same shape as ni!
    ni : float, list or 1D-array [:math:`cm^{-3}`]
        Density of ions corresponding to the atom under consideration. This must
        have the same shape as N1!
    ne : float, list or 1D-array [:math:`cm^{-3}`]
        Electron density to evaluate atomic rates at.
    Te : float, list or 1D-array [:math:`eV`]
        Electron temperature to evaluate atomic rates at.
    rad_prof : list, 1D array or None
        If None, excited state densities are evaluated at all the combinations of ne,Te and zip(Ni,ni).
        If a 1D array (same length as ne,Te,ni and N1), then this is taken to be a radial coordinate
        for radial profiles of ne,Te,ni and N1.
    rad_label : str
        When rad_prof is not None, this is the label for the radial coordinate.
    plot : bool
        Display the excited state ratio

    Returns
    -------
    Nm : array of shape [len(ni)=len(N1),len(ne),len(Te)]
        Density of electrons in excited state `n`  [:math:`cm^{-3}`]
    """
    if m < 1:
        raise ValueError('Excited state principal quantum number must be greater than 1!')
    if m > 9:
        raise ValueError('Selected excited state value not available!')

    ne = np.atleast_1d(ne)
    Te = np.atleast_1d(Te)
    ni = np.atleast_1d(ni)
    N1 = np.atleast_1d(N1)

    if rad_prof is not None:
        # if a radial profile is being computed, ne, Te, ni and N1 must all have the same length
        assert len(ne) == len(Te) and len(ne) == len(ni) and len(ne) == len(N1)

    # get CR model results:
    atom = ehr5_file()

    # coupling to the ground state:
    ground_coupling = atom.res['n{}i_n1'.format(m)]
    cont_coupling = atom.res['n{}ii_n1'.format(m)]

    if rad_prof is not None:
        # evaluate along radial profiles
        gc_interp = interp2d(atom.ne, atom.Te, ground_coupling.T)
        gc = np.array([float(gc_interp(XX, YY)) for XX, YY in zip(ne, Te)])
        cc_interp = interp2d(atom.ne, atom.Te, cont_coupling.T)
        cc = np.array([float(cc_interp(XX, YY)) for XX, YY in zip(ne, Te)])
    else:
        # evaluate at all combinations of points
        gc = interp2d(atom.ne, atom.Te, ground_coupling.T)(ne, Te).T
        cc = interp2d(atom.ne, atom.Te, cont_coupling.T)(ne, Te).T

        N1 = np.rollaxis(np.tile(N1, (cc.shape[0], cc.shape[1], 1)), axis=2)
        ni = np.rollaxis(np.tile(ni, (cc.shape[0], cc.shape[1], 1)), axis=2)
        gc = np.tile(gc, (len(N1), 1, 1))
        cc = np.tile(cc, (len(N1), 1, 1))

    # combine coupling to ground state and to continuum
    Nm = gc * N1 + cc * ni
    Nm_ground = gc * N1
    Nm_cont = cc * ni

    if plot:
        # plot only the first provided value of value of N1 and ni
        ls_style = plot_tools.get_ls_cycle()

        if rad_prof is not None:
            fig, ax = plt.subplots()
            ax.plot(rad_prof, Nm / N1, next(ls_style), lw=2)
            ax.set_ylabel(r'$N_{}/N_1$'.format(m), fontsize=16)
            ax.set_xlabel(rad_label, fontsize=16)
        else:
            fig, ax = plt.subplots(1, 2, figsize=(15, 8))

            labels_Te = ['{:.2e}'.format(val) + ' eV' for val in Te]
            for jt in np.arange(len(Te)):
                ax[0].semilogx(ne, Nm[0, :, jt] / N1[0], next(ls_style), lw=2, label=labels_Te[jt])
            ax[0].legend(loc='best')
            ax[0].set_title('$N_1$=%.2e m$^{-3}$, $n_i$=%.2e m$^{-3}$' % (N1[0], ni[0]))
            ax[0].set_ylabel(r'$N_{}/N_1$'.format(m), fontsize=16)
            ax[0].set_xlabel(r'$n_e$ [cm$^{-3}$]', fontsize=16)

            labels_ne = ['{:.2e}'.format(val) + ' cm$^{-3}$' for val in ne]
            for jn in np.arange(len(ne)):
                ax[1].semilogx(Te, Nm[0, jn, :] / N1[0], next(ls_style), lw=2, label=labels_ne[jn])
            ax[1].legend(loc='best')
            ax[1].set_title('$N_1$=%.2e m$^{-3}$, $n_i$=%.2e m$^{-3}$' % (N1[0], ni[0]))

            ax[1].set_xlabel(r'$T_e$ [eV]', fontsize=16)

        plt.tight_layout()

    return Nm, Nm_ground, Nm_cont



def plot_exc_ratios(n_list=[2, 3, 4, 5, 6, 7, 8, 9], ne=1e13, ni=1e13, Te=50, N1=1e12, 
                    ax=None, ls='-', c='r', label=None):
    """Plot :math:`N_i/N_1`, the ratio of hydrogen neutral density in the excited state `i`
    and the ground state, for given electron density and temperature.

    Parameters
    ----------
    n_list : list of integers
        List of excited states (principal quantum numbers) to consider.
    ne : float
        Electron density in :math:`cm^{-3}`.
    ni : float
        Ionized hydrogen density [:math:`cm^{-3}`]. This may be set equal to ne for a pure plasma.
    Te : float
        Electron temperature in :math:`eV`.
    N1 : float
        Density of ground state hydrogen [:math:`cm^{-3}`]. This is needed because the excited
        state fractions depend on the balance of excitation from the ground state and
        coupling to the continuum.
    ax : matplotlib.axes instance, optional
        Axes instance on which results should be plotted. 
    ls : str
        Line style to use
    c : str or other matplotlib color specification
        Color to use in plots
    label : str
        Label to use in scatter plot. 

    Returns
    -------
    Ns : list of arrays
        List of arrays for each of the n-levels requested, each containing excited state 
        densities at the chosen densities and temperatures for the given ground state density.
    """

    Ns = np.zeros(len(n_list))
    for i, n in enumerate(n_list):
        Ns[i], _, _ = get_exc_state_ratio(m=n, N1=N1, ni=ne, ne=ne, Te=Te, plot=False)

    if ax is None:
        fig, ax = plt.subplots()

    ax.scatter(n_list, Ns / N1, c=c, label=label, s=50.0)
    ax.set_xlabel('n', fontsize=16)
    ax.set_ylabel(r'$N_i/N_1$', fontsize=16)
    ax.set_ylim([0, np.max(Ns / N1) * 1.1])

    return Ns




def Lya_to_neut_dens(emiss_prof, ne, Te, ni=None, plot=True, rhop=None,
                     rates_source='adas', axs=None):
    ''' Estimate ground state neutral density from measured emissivity profiles.
    This ignores possible molecular dynamics and effects that may be captured via
    forward modeling of neutral transport.

    Parameters
    ----------
    emiss_prof : 1D array
        Emissivity profile, units of :math:`W/cm`
    ne : 1D array
        Electron density, units of :math:`cm^{-3}`
    Te : 1D array
        Electron temperature, units of :math:`eV`
    ni : 1D array
        Main ion (H/D/T) density, units of :math:`cm^{-3}`. 
        If left to None, this is internally set to ni=ne.
    plot : bool
        If True, plot some of the key density profiles.
    rhop : 1D array
        Sqrt of normalized poloidal flux radial coordinate. Used only for plotting.
    rates_source : str
        Source of atomic rates. Possible choices are 'adas' or 'colrad'
    axs : Axes instance
        If given, plot on these axes.

    Returns
    ------------
    N1 : 1D array
        Radial profile of estimated ground state atomic neutral density on the same grid 
        as the input arrays. Units of :math:`cm^{-3}`.

    Examples
    --------------
    
    >>> N2_colrad,axs = Lya_to_neut_dens_basic(emiss_prof, ne, Te, ni, 
    >>>                                  plot=True, rhop=rhop, rates_source='colrad')
    
    >>> N2_adas,axs = Lya_to_neut_dens_basic(
    >>>                       emiss_prof, ne, Te, ni, plot=True, rhop=rhop, 
    >>>                       rates_source='adas',axs=axs)
    
    '''
    assert len(emiss_prof)==len(ne) and len(ne)==len(Te)
    if ni is None:
        ni = copy.deepcopy(ne)
    else:
        assert len(ne)==len(ni)

    thirteenpointsix = h*c_light*Rydberg/e
    A_21 = 4.699E8   # s^-1
    E_21 = thirteenpointsix *(1.0 - 2.**(-2.)) * e # J

    # n=2 population in cm^-3
    N2 = emiss_prof/(A_21 * E_21)

    if rates_source=='colrad':
        # use rates from COLRAD code
        atom = ehr5_file()
        ground_coupling = atom.res['n2i_n1']
        cont_coupling = atom.res['n2ii_n1']

        # atomic data is on a grid in units of cm^-3, eV
        gc_interp = interp2d(atom.ne,atom.Te,ground_coupling.T)
        gc = np.array([float(gc_interp(XX,YY)) for XX,YY in zip(ne,Te)])  # Nm/N1 from exc
        cc_interp = interp2d(atom.ne,atom.Te,cont_coupling.T)
        cc = np.array([float(cc_interp(XX,YY)) for XX,YY in zip(ne,Te)])  # Nm/N1 from recomb

        # ground state density
        #N1 = (N2 - cc * ni)/gc
        N1 = N2/gc # continuum recomb term not well constrained, but should be small. Ignore it


    elif rates_source=='adas':
        path = '/home/sciortino/atomAI/atomdat_master/adf15/h/pju#h0.dat'
        log10pec_dict = radiation.read_adf15(path)[1215.2]

        # evaluate these interpolations on our profiles
        pec_recomb = 10**log10pec_dict['recom'].ev(np.log10(ne), np.log10(Te))
        pec_exc = 10**log10pec_dict['excit'].ev(np.log10(ne), np.log10(Te))

        N1 = emiss_prof/E_21/(ne*pec_exc+ni*pec_recomb)


    if plot:
        if rhop is None:
            print('No rhop array was given!')
            rhop = np.arange(len(N1))
        if axs is None:
            fig,ax = plt.subplots(2,1, figsize=(10,7))
        else:
            ax = axs
        sel = np.argmin(np.abs(rhop - 0.9))
        if axs is None:
            # plot this only the first time
            ax[0].plot(rhop[sel:], N2[sel:], label=r'$N_2$')
        ax[0].semilogy(rhop[sel:], N1[sel:], label=fr'{rates_source} $N_1$')
        ax[1].semilogy(rhop[sel:], N1[sel:]/ne[sel:], label=fr'{rates_source} $N_1/n_e$')
        ax[0].set_ylim([1e10,None])
        ax[1].set_ylim([1e-5,None])
        ax[0].set_ylabel(r'Neutral density [$cm^{{-3}}$]')
        ax[1].set_ylabel(r'Density ratio')
        ax[0].legend()
        ax[1].legend()
        ax[1].set_xlabel(r'$\rho_p$')
    else:
        ax = None
        
    return N1,ax




r"""
Aurora functionality to set up and run KN1D to extract atomic and neutral 
background densities at the edge. 

KN1D is a 1D kinetic neutral code originally developed by B.LaBombard (MIT). 
For information, refer to the `KN1D manual  <https://github.com/fsciortino/kn1d/blob/master/kn1d_manual.pdf>`__ .

Note that this Aurora module is merely a wrapper of KN1D. Users require an
IDL license on the computer where this module is called in order to be able to 
run KN1D. The IDL (and Fortran) code themselves are automatically downloaded and
compiled by this module.

sciortino, January 2021
"""

from scipy.interpolate import interp1d
import numpy as np
import os
import scipy.io
from scipy.integrate import cumtrapz
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
from scipy.constants import e, h, c as c_light, Rydberg

from . import neutrals
from . import coords

thisdir = os.path.dirname(os.path.realpath(__file__))

def _setup_kin_profs(rhop, ne_cm3_in, Te_eV_in, Ti_eV_in, 
                     geqdsk, bound_sep_cm, lim_sep_cm,
                     kin_prof_exp_decay_SOL=False, kin_prof_exp_decay_LS=False,
                     ne_decay_len_cm=1.0, Te_decay_len_cm=1.0, Ti_decay_len_cm=1.0, 
                     ne_min_cm3=1e12, Te_min_eV=1.0, Ti_min_eV=1.0):
    '''Private method to set up kinetic profiles to the format required by 
    :py:fun:`~aurora.kn1d.run_kn1d`. Refer to this function for descriptions of inputs.
    
    This function returns ne, Te and Ti profiles on the rmid_to_wall_cm radial grid, 
    from the core to the wall. 

    Parameters
    ----------
    rhop : 1D array
        Sqrt of poloidal flux grid on which ne_cm3, Te_eV and Ti_eV are given.
    ne_cm3_in : 1D array
        Electron density on rhop grid [:math:`cm^{-3}`].
    Te_eV_in : 1D array
        Electron temperature on rhop grid [:math:`eV`].
    Ti_eV_in : 1D array
        Main ion temperature on rhop grid [:math:`eV`].
    geqdsk : `omfit_classes.omfit_eqdsk.OMFITgeqdsk` class instance
        gEQDSK file as processed by the `omfit_classes.omfit_eqdsk.OMFITgeqdsk` class.
    bound_sep_cm : float
        Distance between the wall/boundary and the separatrix [:math:`cm`].
    lim_sep_cm : float
        Distance between the limiter and the separatrix [:math:`cm`].
    kin_prof_exp_decay_SOL : bool
        If True, kinetic profiles are set to exponentially decay over the SOL region.
    kin_prof_exp_decay_LS : bool
        If True, kinetic profiles are set to exponentially decay over the LS region.
    ne_decay_len_cm : list of 2 float
        Exponential decay lengths of electron density in the SOL and LS regions. 
        Default is [1,1] :math:`cm`.
    Te_decay_len_cm : float
        Exponential decay lengths of electron temperature in the SOL and LS regions. 
        Default is [1,1] :math:`cm`.    
    Ti_decay_len_cm : float
        Exponential decay lengths of main ion temperature in the SOL and LS regions. 
        Default is [1,1] :math:`cm`.
    ne_min_cm3 : float
        Minimum electron density across profile. Default is :math:`10^{12} cm^{-3}`.
    Te_min_eV : float
        Minimum electron temperaure across profile. Default is :math:`eV`.
    Ti_min_eV : float
        Minimum main ion temperaure across profile. Default is :math:`eV`.

    Returns
    -------
    rmid_to_wall_cm : 1D array
        Midradius coordinate from the magnetic axis to the wall. Units of [:math:`cm`].
    ne_cm3 : 1D array
        Electron density [:math:`cm^{-3}`] on the rmid_to_wall_cm grid.
    Te_eV : 1D array
        Electron temperature [:math:`eV`] on the rmid_to_wall_cm grid.
    Ti_eV : 1D array
        Main ion temperature [:math:`eV`] on the rmid_to_wall_cm grid.
    '''

    # convert radial coordinate to rmid
    rmid = coords.rad_coord_transform(rhop, 'rhop', 'rmid', geqdsk) # m

    # define radial regions in the SOL in coordinates centered on the mag axis
    rsep = coords.rad_coord_transform(1.0, 'rhop', 'rmid', geqdsk)
    rwall = rsep + bound_sep_cm * 1e-2  # cm-->m
    rlim = rsep + lim_sep_cm * 1e-2  # cm-->m

    # interpolate profiles on grid extending to wall
    rmid_to_wall_m = np.linspace(np.min(rmid), rwall, 1001)  # 101) #201)
    _ne_cm3 = interp1d(rmid, ne_cm3_in, bounds_error=False)(rmid_to_wall_m)  # extrapolates to nan
    _Te_eV = interp1d(rmid, Te_eV_in, bounds_error=False)(rmid_to_wall_m)  # extrapolates to nan
    _Ti_eV = interp1d(rmid, Ti_eV_in, bounds_error=False)(rmid_to_wall_m)  # extrapolates to nan

    indLCFS = np.searchsorted(rmid_to_wall_m, rsep)
    indLS = np.searchsorted(rmid_to_wall_m, rlim)
    ind_end = np.searchsorted(rmid_to_wall_m, rmid[-1])

    # if kinetic profiles don't extend far enough in radius, we must set an exp decay depending on the radial region
    if ind_end < indLS:
        # decays in SOL (all the way to the wall)
        ne_cm3_sol = _ne_cm3[ind_end-1] * np.exp(-(rmid_to_wall_m[ind_end:]-rmid_to_wall_m[ind_end-1])/(ne_decay_len_cm[0]/100.0))
        ne_cm3_ = np.concatenate((_ne_cm3[:ind_end], ne_cm3_sol))
        Te_eV_sol = _Te_eV[ind_end-1] * np.exp(-(rmid_to_wall_m[ind_end:]-rmid_to_wall_m[ind_end-1])/(Te_decay_len_cm[0]/100.0))
        Te_eV_ = np.concatenate((_Te_eV[:ind_end], Te_eV_sol))
        Ti_eV_sol = _Ti_eV[ind_end-1] * np.exp(-(rmid_to_wall_m[ind_end:]-rmid_to_wall_m[ind_end-1])/(Ti_decay_len_cm[0]/100.0))
        Ti_eV_ = np.concatenate((_Ti_eV[:ind_end], Ti_eV_sol))
    else:
        ne_cm3_ = copy.deepcopy(_ne_cm3)
        Te_eV_ = copy.deepcopy(_Te_eV)
        Ti_eV_ = copy.deepcopy(_Ti_eV)

    if ind_end < len(rmid_to_wall_m):
        # decays in the LS
        ne_cm3_ls = ne_cm3_[ind_end-1] * np.exp(-(rmid_to_wall_m[ind_end:]-rmid_to_wall_m[ind_end-1])/(ne_decay_len_cm[1]/100.0))
        ne_cm3 = np.concatenate((ne_cm3_[:ind_end], ne_cm3_ls))
        Te_eV_ls = Te_eV_[ind_end-1] * np.exp(-(rmid_to_wall_m[ind_end:]-rmid_to_wall_m[ind_end-1])/(Te_decay_len_cm[1]/100.0))
        Te_eV = np.concatenate((Te_eV_[:ind_end], Te_eV_ls))
        Ti_eV_ls = Ti_eV_[ind_end-1] * np.exp(-(rmid_to_wall_m[ind_end:]-rmid_to_wall_m[ind_end-1])/(Ti_decay_len_cm[1]/100.0))
        Ti_eV = np.concatenate((Ti_eV_[:ind_end], Ti_eV_ls))
    else:
        ne_cm3 = copy.deepcopy(ne_cm3_)
        Te_eV = copy.deepcopy(Te_eV_)
        Ti_eV = copy.deepcopy(Ti_eV_)


    # User may want to set exp decay in SOL or LS in place of dubious experimental data
    if kin_prof_exp_decay_SOL:
        # decays in the SOL
        ne_cm3[indLCFS:indLS] = ne_cm3[indLCFS - 1] * np.exp(
            -(rmid_to_wall_m[indLCFS:indLS] - rmid_to_wall_m[indLCFS-1])/(ne_decay_len_cm[0]/100.0)
        )
        Te_eV[indLCFS:indLS] = Te_eV[indLCFS - 1] * np.exp(
            -(rmid_to_wall_m[indLCFS:indLS] - rmid_to_wall_m[indLCFS-1])/(Te_decay_len_cm[0]/100.0)
        )
        Ti_eV[indLCFS:indLS] = Ti_eV[indLCFS - 1] * np.exp(
            -(rmid_to_wall_m[indLCFS:indLS] - rmid_to_wall_m[indLCFS-1])/(Ti_decay_len_cm[0]/100.0)
        )

    if kin_prof_exp_decay_LS:
        # decays in the LS
        ne_cm3[indLS:] = ne_cm3[indLS-1] * np.exp(-(rmid_to_wall_m[indLS:]-rmid_to_wall_m[indLS-1])/(ne_decay_len_cm[1]/100.0))
        Te_eV[indLS:] = Te_eV[indLS-1] * np.exp(-(rmid_to_wall_m[indLS:]-rmid_to_wall_m[indLS-1])/(Te_decay_len_cm[1]/100.0))
        Ti_eV[indLS:] = Ti_eV[indLS-1] * np.exp(-(rmid_to_wall_m[indLS:]-rmid_to_wall_m[indLS-1])/(Ti_decay_len_cm[1]/100.0))
    
    # set minima across radial profiles
    ne_cm3[ne_cm3 < ne_min_cm3] = ne_min_cm3
    Te_eV[Te_eV < Te_min_eV] = Te_min_eV
    Ti_eV[Ti_eV < Ti_min_eV] = Ti_min_eV

    return rmid_to_wall_m, ne_cm3, Te_eV, Ti_eV




def run_kn1d(rhop, ne_cm3, Te_eV, Ti_eV, geqdsk, p_H2_mTorr, 
             clen_divertor_cm, clen_limiter_cm, bound_sep_cm, lim_sep_cm,
             innermost_rmid_cm=5.0, mu=2., pipe_diag_cm=0.0, vx=0.0, collisions={},
             kin_prof_exp_decay_SOL=False, kin_prof_exp_decay_LS=False,
             ne_decay_len_cm=[1.0,1.0], Te_decay_len_cm=[1.0,1.0], Ti_decay_len_cm=[1.,1.], 
             ne_min_cm3=1e12, Te_min_eV=1.0, Ti_min_eV=1.0, plot_kin_profs=False):
    '''Run KN1D for the given parameters. Refer to the KN1D manual for details. 

    Depending on the provided options, kinetic profiles are extended beyond the Last Closed
    Flux Surface (LCxFS) and the Limiter Shadow (LS) via exponential decays with specified 
    decay lengths. It is assumed that the given kinetic profiles extend from the core until
    at least the LCFS. All inputs are taken to be time-independent.

    This function automatically checks if a KN1D repository is available; if it is not,
    it obtains it from the web and compiles the necessary code. 

    Note that an IDL license must be available. Aurora does not currently include a Python
    translation of KN1D -- it only acts as a wrapper.

    Parameters
    ----------
    rhop : 1D array
        Sqrt of poloidal flux grid on which ne_cm3, Te_eV and Ti_eV are given.
    ne_cm3 : 1D array
        Electron density on rhop grid [:math:`cm^{-3}`].
    Te_eV : 1D array
        Electron temperature on rhop grid [:math:`eV`].
    Ti_eV : 1D array
        Main ion temperature on rhop grid [:math:`eV`].
    geqdsk : `omfit_classes.omfit_eqdsk.OMFITgeqdsk` class instance
        gEQDSK file as processed by the `omfit_classes.omfit_eqdsk.OMFITgeqdsk` class.
    p_H2_mTorr : float
        Pressure of molecular hydrogen-isotopes measured at the wall. This may be estimated
        from experimental pressure gauges. This variable effectively sets the amplitude of the 
        neutral source at the edge. Units of :math:`mTorr`.
    clen_divertor_cm : float
        Connection length from the midplane to the divertor [:math:`cm`].
    clen_limiter_cm : float
        Connection length from the midplane to the limiter [:math:`cm`].
    bound_sep_cm : float
        Distance between the wall/boundary and the separatrix [:math:`cm`].
    lim_sep_cm : float
        Distance between the limiter and the separatrix [:math:`cm`].
    innermost_rmid_cm : float
        Distance from the wall to solve for. Default is 5 cm. 
    mu : float
        Atomic mass number of simulated species. Default is 2.0 (D). 
    pipe_diag_cm : float
        Diameter of the pipe through which H2 pressure is measured (see `p_H2_mTorr` variable). 
        If left to 0, this diameter is effectively set to infinity. Default is 0. 
    vx : float
        Radial velocity imposed on neutrals. This only has a weak effect usually. 
        Default is 0 [:math:`cm/s`].
    collisions : dict
        Collision terms flags. Set each to True or False. If any of the flags are not given, 
        all collision terms are internally set to be active. Possible flags are
        'H2_H2_EL','H2_P_EL','H2_H_EL','H2_HP_CX','H_H_EL','H_P_CX','H_P_EL','Simple_CX'
    kin_prof_exp_decay_SOL : bool
        If True, kinetic profiles are set to exponentially decay over the SOL region.
    kin_prof_exp_decay_LS : bool
        If True, kinetic profiles are set to exponentially decay over the LS region.
    ne_decay_len_cm : list of 2 float
        Exponential decay lengths of electron density in the SOL and LS regions. 
        Default is [1,1] :math:`cm`.
    Te_decay_len_cm : float
        Exponential decay lengths of electron temperature in the SOL and LS regions. 
        Default is [1,1] :math:`cm`.    
    Ti_decay_len_cm : float
        Exponential decay lengths of main ion temperature in the SOL and LS regions. 
        Default is [1,1] :math:`cm`.
    ne_min_cm3 : float
        Minimum electron density across profile. Default is :math:`10^{12} cm^{-3}`.
    Te_min_eV : float
        Minimum electron temperaure across profile. Default is :math:`eV`.
    Ti_min_eV : float
        Minimum main ion temperaure across profile. Default is :math:`eV`.
    plot_kin_profs : bool
        If True, kinetic profiles input to KN1D are plotted.
    
    Returns
    -------
    dict
        KN1D results and inputs, all collected into a dictionary. See example script for
        an illustration of using this. 

    Notes
    -----
    For an example application, see the examples/aurora_kn1d.py script. 
    '''
    
    if 'IDL_STARTUP' not in os.environ:
        raise ValueError('An IDL installation does not seem to be available! KN1D cannot be run.')

    cwd = os.getcwd()
    
    # make sure that the KN1D source code is accessible. 
    if 'KN1D' not in os.listdir(thisdir):
        #if 'KN1D_DIR' not in os.environ:
        # git clone the KN1D repository
        os.system(f'git clone https://github.com/fsciortino/kn1d {thisdir}/KN1D')

        os.chdir(f'{thisdir}/KN1D')
        # compile fortran libraries
        print(f'export KN1D_DIR={thisdir}/KN1D; make clean; make')
        os.system(f'export KN1D_DIR={thisdir}/KN1D; make clean; make')
        os.chdir(cwd)
        #else:
        # copy KN1D directory locally
        #shutil.copytree(os.environ['KN1D_DIR'],thisdir+'/KN1D')
    else:
        # KN1D directory already available, assumed to be already built
        # NB: users need to have write-access to this directory!
        pass

    kn1d={}
    kn1d.update(collisions)
    if 'H2_H2_EL' not in kn1d:
        kn1d['H2_H2_EL'] = True
    if 'H2_P_EL' not in kn1d:
        kn1d['H2_P_EL'] = True
    if 'H2_H_EL' not in kn1d:
        kn1d['H2_H_EL'] = True
    if 'H2_HP_CX' not in kn1d:
        kn1d['H2_HP_CX'] = True
    if 'H_H_EL' not in kn1d:
        kn1d['H_H_EL'] = True
    if 'H_P_CX' not in kn1d:
        kn1d['H_P_CX'] = True
    if 'H_P_EL' not in kn1d:
        kn1d['H_P_EL'] = True
    if 'Simple_CX' not in kn1d:
        kn1d['Simple_CX'] = False

    # get kinetic profiles on rmid_to_wall_m grid, applying exponential decays in SOL as requested
    rmid_to_wall_m,_ne_cm3,_Te_eV,_Ti_eV = _setup_kin_profs(rhop, ne_cm3, Te_eV, Ti_eV, 
                                              geqdsk, bound_sep_cm, lim_sep_cm,
                                              kin_prof_exp_decay_SOL, kin_prof_exp_decay_LS,
                                              ne_decay_len_cm, Te_decay_len_cm, Ti_decay_len_cm, 
                                              ne_min_cm3, Te_min_eV, Ti_min_eV)

    if plot_kin_profs:
        # show kinetic profiles going into KN1D modeling
        plot_input_kin_prof(rmid_to_wall_m, _ne_cm3, _Te_eV, _Ti_eV,
                            innermost_rmid_cm, bound_sep_cm, lim_sep_cm)

    rhop = coords.rad_coord_transform(rmid_to_wall_m, 'rmid', 'rhop', geqdsk)
    rwall_m = rmid_to_wall_m[-1]  #m
    rsep_m = coords.rad_coord_transform(1.0, 'rhop', 'rmid', geqdsk) # m
    rlim_m = rsep_m + lim_sep_cm * 1e-2  # m


    # KN1D defines coordinates from the wall INWARD. Invert now:
    r_kn1d_m = np.abs(rmid_to_wall_m - rwall_m)[::-1]
    rlim_kn1d_m = np.abs(rlim_m - rwall_m)
    rsep_kn1d_m = np.abs(rsep_m - rwall_m)

    # diameter of pressure gauge pipe. Allows collisions with side-walls to be simulated
    dPipe = pipe_diag_cm*np.ones(len(rmid_to_wall_m))*1e-2  # m  -- zero values are treated as infinity

    # define the connection length vector
    lc = np.zeros(len(rmid_to_wall_m))
    lc[(rsep_m < rmid_to_wall_m) * (rmid_to_wall_m < rlim_m)] = clen_divertor_cm * 1e-2  # m
    lc[rmid_to_wall_m > rlim_m] = clen_limiter_cm * 1e-2  # m


    def idl_array(arr, num):
        # set arrays into string form for IDL format
        return '[' + ','.join([str(round(val, num)) for val in arr]) + ']'
    
    def round_arr(arr, num, dtype=str):
        # avoid issues with floating-point precision
        return np.array([round(val, num) for val in arr], dtype=dtype)
    
    # cut all radial profiles to given innermost location
    ind_in = np.searchsorted(r_kn1d_m, innermost_rmid_cm * 1e-2)  # r_kn1d_m is from wall inwards
    
    # Save arrays in final forms for KN1D 
    num = 5  # digital point precision
    kn1d['x'] = round_arr(r_kn1d_m[:ind_in], num, float)
    ne_ = _ne_cm3[::-1]*1e6
    kn1d['ne_m3'] = round_arr(ne_[:ind_in], num, float) # m^-3
    Te_ = _Te_eV[::-1]
    kn1d['Te_eV'] = round_arr(Te_[:ind_in], num, float)  # eV
    Ti_ = _Ti_eV[::-1]
    kn1d['Ti_eV'] = round_arr(Ti_[:ind_in], num, float) # eV
    dPipe_ = dPipe[::-1]
    kn1d['dPipe'] = round_arr(dPipe_[:ind_in], num, float)
    lc_ = lc[::-1]
    kn1d['lc'] = round_arr(lc_[:ind_in], num, float)
    kn1d['xlim'] = rlim_kn1d_m
    kn1d['xsep'] = rsep_kn1d_m
    kn1d['p_H2_mTorr'] =  p_H2_mTorr  # mTorr
    kn1d['mu'] = mu
    

    # finally, set plasma radial velocity profile (negative is towards the wall [m s^-1])
    vx = np.ones_like(kn1d['x']) * vx
    kn1d['vx'] = round_arr(vx, num, float)

    # Create IDL script to run KN1D
    idl_cmd = '''
; Input data for KN1D run
x = {x:}
x_lim = {x_lim:.5f}
xsep = {xsep:.5f}
gaugeH2 = {gaugeH2:}
mu = {mu:}
Ti = {Ti:}
Te = {Te:}
dens = {ne:}
vx = {vx:}
lc = {lc:}
dPipe = {dPipe:}

; collisions options are set via a common block
common KN1D_collisions,H2_H2_EL,H2_P_EL,H2_H_EL,H2_HP_CX,H_H_EL,H_P_EL,H_P_CX,Simple_CX

H2_H2_EL= {H2_H2_EL:}
H2_P_EL = {H2_P_EL:}
H2_H_EL = {H2_H_EL:}
H_H_EL= {H_H_EL:}
H_P_CX = {H_P_CX:}
H_P_EL = {H_P_EL:}
Simple_CX = {Simple_CX:}

; now run KN1D
kn1d,x,x_lim,xsep,gaugeH2,mu,Ti,Te,dens,vx,lc,dPipe, xh2,nh2,gammaxh2,th2,qxh2_total,nhp,thp,sh,sp, xh,nh,gammaxh,th,qxh_total,nethsource,sion,qh_total,sidewallh,lyman,balmer,gammahlim

; save result to an IDL .sav file
save,xh2,nh2,gammaxh2,th2,qxh2_total,nhp,thp,sh,sp, xh,nh,gammaxh,th,qxh_total,nethsource,sion,qh_total,sidewallh,lyman,balmer,gammahlim,filename = "kn1d_out.sav"

exit

    '''.format(
        x=idl_array(kn1d['x'], num),
        x_lim=kn1d['xlim'],
        xsep=kn1d['xsep'],
        gaugeH2=kn1d['p_H2_mTorr'],
        mu=kn1d['mu'],
        Ti=idl_array(kn1d['Ti_eV'], num),
        Te=idl_array(kn1d['Te_eV'], num),
        ne=idl_array(kn1d['ne_m3'], num),
        vx=idl_array(kn1d['vx'], num),
        lc=idl_array(kn1d['lc'], num),
        dPipe=idl_array(kn1d['dPipe'], num),
        H2_H2_EL=int(kn1d['H2_H2_EL']),
        H2_P_EL=int(kn1d['H2_P_EL']),
        H2_H_EL=int(kn1d['H2_H_EL']),
        H_H_EL=int(kn1d['H_H_EL']),
        H_P_CX=int(kn1d['H_P_CX']),
        H_P_EL=int(kn1d['H_P_EL']),
        Simple_CX=int(kn1d['Simple_CX']),
    )

    # write IDL file
    with open(f'{thisdir}/KN1D/new_kn1d_run.pro', 'w') as f:
        f.write(idl_cmd)

    # Run the script 
    os.system(f'cd {thisdir}/KN1D; idl new_kn1d_run.pro')

    #### store all KN1D data for postprocessing  #####
    res = {}
    #res['ins'] = kn1d  # store inputs for plotting
    out = res['out'] = scipy.io.readsav(f'{thisdir}/KN1D/kn1d_out.sav')
    res['kn1d_input'] = scipy.io.readsav(f'{thisdir}/KN1D/.KN1D_input')
    res['kn1d_mesh'] = scipy.io.readsav(f'{thisdir}/KN1D/.KN1D_mesh')
    res['kn1d_H2'] = scipy.io.readsav(f'{thisdir}/KN1D/.KN1D_H2')
    res['kn1d_H'] = scipy.io.readsav(f'{thisdir}/KN1D/.KN1D_H')

    os.chdir(cwd)

    # ---------------------------
    # Additional processed outputs
    # Compute ion flux by integrating over atomic ionization rate
    Sion = out['sion']
    Sion_interp = interp1d(out['xh'], Sion, bounds_error=False, fill_value=0.0)(kn1d['x'])
    out['Gamma_i'] = cumtrapz(Sion_interp, kn1d['x'], initial=0.0)

    # Effective diffusivity
    out['D_eff'] = np.abs(out['Gamma_i'] / np.gradient(kn1d['ne_m3'], kn1d['x']))  # check

    # ensure that x bases are all to the same accuracy to avoid issues in interpolation
    out['xh'] = round_arr(out['xh'], num, dtype=float)
    out['xh2'] = round_arr(out['xh2'], num, dtype=float)

    # gradient length scales
    out['L_ne'] = np.abs(1.0 / np.gradient(np.log(kn1d['ne_m3']), kn1d['x']))
    out['L_Te'] = np.abs(1.0 / np.gradient(np.log(kn1d['Te_eV']), kn1d['x']))
    out['L_Ti'] = np.abs(1.0 / np.gradient(np.log(kn1d['Ti_eV']), kn1d['x']))

    #### Calculate radial profiles of neutral excited states   ####

    # take KN1D neutral density profile to be the ground state (excited states are a small correction wrt this)
    N1 = interp1d(out['xh'], out['nh'], kind='linear')(kn1d['x'])

    # assume pure plasma and quasi-neutrality
    nhp_interp = interp1d(out['xh2'], out['nhp'], bounds_error=False, fill_value=0.0)(kn1d['x'])  # nH2+
    out['ni'] = kn1d['ne_m3'] - nhp_interp

    # get profiles of excited states' density (only n=2 and n=3) ---- densities in cm^-3, Te in eV
    N2, N2_ground, N2_cont = neutrals.get_exc_state_ratio(
        m=2, N1=N1/1e6, ni=out['ni']/1e6, ne=kn1d['ne_m3'] / 1e6, Te=kn1d['Te_eV'], plot=False, rad_prof=kn1d['x']
    )
    N3, N3_ground, N3_cont = neutrals.get_exc_state_ratio(
        m=3, N1=N1/1e6, ni=out['ni']/1e6, ne=kn1d['ne_m3'] / 1e6, Te=kn1d['Te_eV'], plot=False, rad_prof=kn1d['x']
    )

    out['N2'] = N2 * 1e6  # transform back to m^-3 (all KN1D units are in SI units)
    out['N2_ground'] = N2_ground * 1e6
    out['N2_cont'] = N2_cont * 1e6
    out['N3'] = N3 * 1e6
    out['N3_ground'] = N3_ground * 1e6
    out['N3_cont'] = N3_cont * 1e6

    # store rwall_m to allow external coordinate conversions
    out['rwall_m'] = rwall_m
    
    #################################
    # Store neutral density profiles in a format that can be used for integrated modeling
    out_profs = res['kn1d_profs'] = {}

    # save profiles on (inverted) grid extending all the way to the axis (extrapolating)
    _rhop = coords.rad_coord_transform(rwall_m - kn1d['x'][::-1], 'rmid', 'rhop', geqdsk)
    out_profs['rhop'] = np.linspace(0.0, 1.1, 200)

    out_profs['n0'] = np.exp(interp1d(_rhop, np.log(N1[::-1]), 
                                      bounds_error=False, fill_value='extrapolate')(out_profs['rhop']))
    out_profs['n0_n2'] = np.exp(interp1d(_rhop, np.log(out['N2'][::-1]), 
                                         bounds_error=False, fill_value='extrapolate')(out_profs['rhop']))
    out_profs['n0_n3'] = np.exp(interp1d(_rhop, np.log(out['N3'][::-1]), 
                                         bounds_error=False, fill_value='extrapolate')(out_profs['rhop']))

    # also save profiles of Ly- and H/D-alpha
    _rhop_emiss = coords.rad_coord_transform(rwall_m - out['xh'][::-1], 'rmid', 'rhop', geqdsk)
    out_profs['lyman'] = np.exp(
        interp1d(_rhop_emiss, np.log(out['lyman'][::-1]), 
                 bounds_error=False, fill_value='extrapolate')(out_profs['rhop'])
    )
    out_profs['balmer'] = np.exp(
        interp1d(_rhop_emiss, np.log(out['balmer'][::-1]), 
                 bounds_error=False, fill_value='extrapolate')(out_profs['rhop'])
    )
        
    return res






    
def plot_input_kin_prof(rmid_to_wall_m, ne_cm3, Te_eV, Ti_eV,
                        innermost_rmid_cm, bound_sep_cm, lim_sep_cm):
    '''Plot extent of kinetic profiles entering KN1D calculation
    '''
    fig,axs = plt.subplots(3,1, figsize=(8,8), sharex=True)
    axs[0].plot(rmid_to_wall_m*1e2, ne_cm3)
    axs[1].plot(rmid_to_wall_m*1e2, Te_eV)
    axs[2].plot(rmid_to_wall_m*1e2, Ti_eV)
    axs[-1].set_xlabel(r'$r_{mid}$ [cm]')
    axs[0].set_ylabel(r'$n_e$ [$cm^{-3}$]')
    axs[1].set_ylabel(r'$T_e$ [$eV$]')
    axs[2].set_ylabel(r'$T_i$ [$eV$]')

    #from IPython import embed
    #embed()
    indin = np.argmin(np.abs(rmid_to_wall_m*100 -(np.max(rmid_to_wall_m)*100.-innermost_rmid_cm)))
    
    axs[0].set_ylim([0,np.max(ne_cm3[indin:])]); axs[0].grid(True)
    axs[1].set_ylim([0,np.max(Te_eV[indin:])]); axs[1].grid(True)
    axs[2].set_ylim([0,np.max(Ti_eV[indin:])]); axs[2].grid(True)
    axs[0].set_xlim([np.max(rmid_to_wall_m)*100.-innermost_rmid_cm, np.max(rmid_to_wall_m)*100.])
    plt.subplots_adjust(hspace=0)

    # also show location of limiter and LCFS:
    axs[0].axvline(np.max(rmid_to_wall_m)*1e2-bound_sep_cm,c='m')
    axs[1].axvline(np.max(rmid_to_wall_m)*1e2-bound_sep_cm,c='m')
    axs[2].axvline(np.max(rmid_to_wall_m)*1e2-bound_sep_cm,c='m')

    axs[0].axvline(np.max(rmid_to_wall_m)*1e2-(bound_sep_cm-lim_sep_cm),c='m')
    axs[1].axvline(np.max(rmid_to_wall_m)*1e2-(bound_sep_cm-lim_sep_cm),c='m')
    axs[2].axvline(np.max(rmid_to_wall_m)*1e2-(bound_sep_cm-lim_sep_cm),c='m')
    







def plot_overview(res):
    '''Plot an overview of a KN1D run, showing both kinetic profile inputs and 
    a small selection of the outputs. 

    Parameters
    ----------
    res : dict
        Output dictionary from function :py:fun:`~aurora.kn1d.run_kn1d`.
    '''
    
    ins = res['kn1d_input']
    outs = res['out']

    fig, ax = plt.subplots(4, 1, sharex=True, figsize=(10, 10))
    
    mu = int(ins['mu'])
    species = 'H' if mu == 1 else 'D'

    (line,) = ax[0].plot(ins['x'], ins['n'] / 1e19, lw=2.0)

    c = line.get_color()

    ax[1].semilogy(ins['x'], ins['te'], lw=2.0, c=c, ls='-', label=r'$T_e$')
    ax[1].semilogy(ins['x'], ins['ti'], lw=2.0, c=c, ls='--', label=r'$T_i$')
    ax[1].semilogy(outs['xh'], outs['th'], lw=2.0, c=c, ls='-.', label=fr'$T_{species}$')

    ax[2].semilogy(outs['xh'], outs['nh'], lw=2.0, c=c, ls='-', label=fr'$n_{species}$')
    ax[2].semilogy(outs['xh2'], outs['nh2'], lw=2.0, c=c, ls='--', label=r'$n_{%s2}$'%species)
    ax[2].semilogy(outs['xh2'], outs['nhp'], lw=2.0, c=c, ls='-.', label=r'$n_{%s2}^+$'%species)
    
    # quasineutrality in a pure plasma: nH+ = ne - nH2+ (NB: NH2+ is saved as nhp)
    nhp_interp = interp1d(outs['xh2'], outs['nhp'], bounds_error=False, fill_value=0.0)(ins['x'])
    ax[2].semilogy(ins['x'], ins['n'] - nhp_interp, lw=2.0, c=c, ls=':', label=r'$n_e - n_{%s2}^+$'%species)

    ax[3].semilogy(outs['xh'], outs['sion'] / 1e20, lw=2.0, c=c, label='Atomic ionization rate')

    # annotate location of limiter and LCFS
    ax[0].axvline(ins['xlimiter'])
    ax[0].axvline(ins['xsep'])
    ax[1].axvline(ins['xlimiter'])
    ax[1].axvline(ins['xsep'])
    ax[2].axvline(ins['xlimiter'])
    ax[2].axvline(ins['xsep'])
    ax[3].axvline(ins['xlimiter'])
    ax[3].axvline(ins['xsep'])

    dist = ins['xlimiter'] / 10.0  # convenient rule-of-thumb for plotting
    ax[0].annotate('Limiter', (ins['xlimiter'] + dist, 0.5 * ax[0].get_ylim()[1]), fontsize=14, rotation=90)
    ax[0].annotate('LCFS', (ins['xsep'] + dist, 0.5 * ax[0].get_ylim()[1]), fontsize=14, rotation=90)
    
    # reduce number of ticks from default
    for n_ax in [0, 1, 2, 3]:
        nyticks = len(ax[n_ax].get_yticks())
        ax[n_ax].set_yticks(ax[n_ax].get_yticks()[::2])
        
    # set every other ylabel/ticks to the right
    ax[1].yaxis.set_label_position("right")
    ax[1].yaxis.tick_right()
    ax[3].yaxis.set_label_position("right")
    ax[3].yaxis.tick_right()

    ax[0].set_ylabel(r'$n_e$ [$10^{19}$ $m^{-3}$]')
    ax[1].set_ylabel(r'$eV$')
    ax[2].set_ylabel(r'$m^{-3}$')
    ax[3].set_ylabel(r'$10^{20}$ $m^{-3}$')
    ax[-1].set_xlabel('Distance from the wall [m]')
    
    # legends
    #ax[0].legend(fontsize=14, loc='best')
    ax[1].legend(fontsize=14, loc='best')
    ax[2].legend(fontsize=14, loc='best')
    ax[3].legend(fontsize=14, loc='best')


def plot_exc_states(res):
    '''Plot excited state fractions of atomic neutral density from a KN1D run.

    Parameters
    ----------
    res : dict
        Output dictionary from function :py:fun:`~aurora.kn1d.run_kn1d`.
    '''
    ins = res['kn1d_input']
    outs = res['out']

    fig, ax = plt.subplots(4, 1, sharex=True, figsize=(10, 10))

    _ne = interp1d(ins['x'], ins['n'],bounds_error=False,fill_value='extrapolate')(outs['xh'])
    (line,) = ax[0].semilogy(outs['xh'], outs['nh'] / _ne, lw=2.0)
    c = line.get_color()

    ax[1].plot(outs['xh'], outs['nh'], c=c, lw=2.0)

    ax[2].plot(ins['x'], outs['N2'], c=c, lw=2.0, ls='-', label='total')
    ax[2].plot(ins['x'], outs['N2_ground'], c=c, lw=2.0, ls='--', label='from ground')
    ax[2].plot(ins['x'], outs['N2_cont'], c=c, lw=2.0, ls='-.', label='from cont.')

    ax[3].plot(ins['x'], outs['N3'], c=c, ls='-', lw=2.0, label='total')
    ax[3].plot(ins['x'], outs['N3_ground'], c=c, lw=2.0, ls='--', label='from ground')
    ax[3].plot(ins['x'], outs['N3_cont'], c=c, lw=2.0, ls='-.', label='from cont.')

    # annotate location of limiter and LCFS
    ax[0].axvline(ins['xlimiter'])
    ax[0].axvline(ins['xsep'])
    ax[1].axvline(ins['xlimiter'])
    ax[1].axvline(ins['xsep'])
    ax[2].axvline(ins['xlimiter'])
    ax[2].axvline(ins['xsep'])
    ax[3].axvline(ins['xlimiter'])
    ax[3].axvline(ins['xsep'])
    
    dist = ins['xlimiter'] / 10.0  # convenient rule-of-thumb for plotting
    ax[0].annotate('Limiter', (ins['xlimiter'] + dist, 1e-3 * ax[0].get_ylim()[1]), fontsize=14, rotation=90)
    ax[0].annotate('LCFS', (ins['xsep'] + dist, 1e-3 * ax[0].get_ylim()[1]), fontsize=14, rotation=90)
    
    ax[0].set_ylabel(r'$n_{n,1}/n_e$')
    #ax[0].legend(loc='best')
    
    ax[1].set_ylabel(r'$n_{n,1}$ [m$^{-3}$]')
    
    ax[2].set_ylabel(r'$n_{n,2}$ [m$^{-3}$]')
    ax[2].legend(loc='best')
    
    ax[3].set_ylabel(r'$n_{n,3}$ [m$^{-3}$]')
    ax[3].legend(loc='best')
    
    ax[-1].set_xlabel('Distance from the wall [m]')



    
def plot_emiss(res, check_collrad=True):
    '''Plot profiles of Ly-a and D-alpha emissivity from the KN1D output. 
    KN1D internally computes Ly-a and D-alpha emission using the Johnson-Hinnov
    coefficients; here we check the result of that calculation and compare it to the 
    prediction from atomic data from the COLLRAD collisional-radiative model included
    in DEGAS2. 

    Parameters
    ----------
    res : dict
        Output dictionary from function :py:fun:`~aurora.kn1d.run_kn1d`.
    check_collrad : bool
        If True, compare KN1D prediction of Ly-a and D-a emission using Johnson-Hinnov
        rates using rates from COLLRAD. 
    '''
    
    ins = res['kn1d_input']
    outs = res['out']

    mu = int(ins['mu'])
    fig, ax = plt.subplots(2, 1, sharex=True, figsize=(10, 8))
    (line,) = ax[0].plot(outs['xh'], outs['lyman'], ls='-')
    c = line.get_color()

    ax[1].plot(outs['xh'], outs['balmer'], '-', c=c, label='KN1D (JH)')

    # annotate location of limiter and LCFS
    ax[0].axvline(ins['xlimiter'])
    ax[0].axvline(ins['xsep'])
    ax[1].axvline(ins['xlimiter'])
    ax[1].axvline(ins['xsep'])

    dist = ins['xlimiter'] / 10.0  # convenient rule-of-thumb for plotting
    ax[0].annotate('Limiter', (ins['xlimiter'] + dist, 0.5 * ax[0].get_ylim()[1]), fontsize=14, rotation=90)
    ax[0].annotate('LCFS', (ins['xsep'] + dist, 0.5 * ax[0].get_ylim()[1]), fontsize=14, rotation=90)

    ax[-1].set_xlabel('Distance from the wall [m]')
    # ---------------
    if check_collrad:
        # test KN1D calculation of Ly-a and D-a

        thirteenpointsix = h * c_light * Rydberg / e
        E_32 = thirteenpointsix * (2.0 ** (-2.0) - 3.0 ** (-2.0)) * e  # J
        E_21 = thirteenpointsix * (1.0 - 2.0 ** (-2.0)) * e  # J

        # Balmer wavelengths from DEGAS2:
        # \lambda_{H_\alpha} ~ 6562.80 A
        # \lambda_{D_\alpha} ~ 6561.04 A
        # \lambda_{T_\alpha} ~ 6560.45 A

        # Lyman series spontaneous emission coeffs for n=2 to 1, 3 to 1, ... 16 to 1
        A_lyman = [4.699e8, 5.575e7, 1.278e7, 4.125e6, 1.644e6, 7.568e5, 3.869e5, 2.143e5,
                   1.263e5, 7.834e4, 5.066e4, 3.393e4, 2.341e4, 1.657e4, 1.200e4]
        
        # Balmer series spontaneous emission coeffs for n=3 to 2, 4 to 2, ... 17 to 2
        A_balmer = [4.41e7, 8.42e6, 2.53e6, 9.732e5, 4.389e5, 2.215e5, 1.216e5, 7.122e4,
                    4.397e4, 2.83e4, 18288.8, 12249.1, 8451.26, 5981.95, 4332.13]

        A_21 = A_lyman[0]  # Ly-alpha
        A_32 = A_balmer[0]  # D-alpha

        Ly_alpha = interp1d(ins['x'], E_21 * outs['N2'] * A_21, kind='linear', bounds_error=False)(outs['xh'])  # J/s/m^3
        D_alpha = interp1d(ins['x'], E_32 * outs['N3'] * A_32, kind='linear', bounds_error=False)(outs['xh'])  # J/s/m^3

        ax[0].plot(outs['xh'], Ly_alpha, '--', c=c)
        ax[1].plot(outs['xh'], D_alpha, '--', c=c, label='collrad')

        ax[0].set_ylabel(r'Ly-alpha [W m$^{-3}$]')
        ax[1].set_ylabel(fr'{"H" if mu == 1 else "D"}-alpha [W m$^{-3}$]')  # 656.28 nm in air
        ax[-1].set_xlabel('Distance from the wall [m]')
        
        ax[0].legend(loc='best')
        ax[1].legend(loc='best')
        






def plot_transport(res):
    '''Make a simple set of plots of gradient scale lengths and effective diffusion coefficients
    from the KN1D output.
    
    Parameters
    ----------
    res : dict
        Output dictionary from function :py:fun:`~aurora.kn1d.run_kn1d`.
    '''
    ins = res['kn1d_input']
    outs = res['out']

    fig, ax = plt.subplots(3, 1, sharex=True, figsize=(10, 7.5))
    
    (line,) = ax[0].plot(ins['x'], outs['L_ne'], lw=2.0, label=r'$L_{n_e}$')
    c = line.get_color()

    ax[0].plot(ins['x'], outs['L_Te'], lw=2.0, c=c, ls='--', label=r'$L_{T_e}$')
    ax[0].plot(ins['x'], outs['L_Ti'], lw=2.0, c=c, ls='-.', label=r'$L_{T_i}$')

    # Effective diffusivity from Gamma_i/\nabla(n_e)
    ax[1].plot(ins['x'], outs['D_eff'], lw=2.0, c=c)

    # atomic and ion source profiles
    ax[2].plot(outs['xH2'], outs['sh'], lw=2.0, c=c, ls='-', label='Atomic source')
    ax[2].plot(outs['xH2'], outs['sp'], lw=2.0, c=c, ls='--', label='Ion source')

    # annotate location of limiter and LCFS
    ax[0].axvline(ins['xlimiter'])
    ax[0].axvline(ins['xsep'])
    ax[1].axvline(ins['xlimiter'])
    ax[1].axvline(ins['xsep'])
    ax[2].axvline(ins['xlimiter'])
    ax[2].axvline(ins['xsep'])
    
    dist = ins['xlimiter'] / 10.0  # convenient rule-of-thumb for plotting
    ax[0].annotate('Limiter', (ins['xlimiter'] + dist, 0.5 * ax[0].get_ylim()[1]), fontsize=14, rotation=90)
    ax[0].annotate('LCFS', (ins['xsep'] + dist, 0.5 * ax[0].get_ylim()[1]), fontsize=14, rotation=90)

    ax[0].legend(fontsize=16, loc='best')
    
    ax[1].set_ylabel(r'$D_{eff}$ [$m^2$ $s^{-1}$]', fontsize=16)
    #ax[1].legend(fontsize=16, loc='best')

    ax[2].set_ylabel(r'$m^{-3}$ $s^{-1}$', fontsize=16)
    ax[2].legend(fontsize=16, loc='best')
    
    ax[-1].set_xlabel('Distance from the wall [m]')
